#
# This file is part of LiteSATA.
#
# Copyright (c) 2015-2019 Florent Kermarrec <florent@enjoy-digital.fr>
# Copyright (c) 2016 Olof Kindgren <olof.kindgren@gmail.com>
# SPDX-License-Identifier: BSD-2-Clause

from collections import OrderedDict
from functools import reduce
from operator import xor

from litesata.common import *

# Link CRC -----------------------------------------------------------------------------------------

class CRCEngine(Module):
    """Cyclic Redundancy Check Engine

    Compute next CRC value from last CRC value and data input using
    an optimized asynchronous LFSR.

    Parameters
    ----------
    width : int
        Width of the data bus and CRC.
    polynom : int
        Polynom of the CRC (ex: 0x04C11DB7 for IEEE 802.3 CRC)

    Attributes
    ----------
    d : in
        Data input.
    last : in
        last CRC value.
    next :
        next CRC value.
    """
    def __init__(self, width, polynom):
        self.data = Signal(width)
        self.last = Signal(width)
        self.next = Signal(width)

        # # #

        def _optimize_xors(l):
            """
            remove an even numbers of XORs with the same bit
            replace an odd number of XORs with a single XOR
            """
            d = OrderedDict()
            for e in l:
                try:
                    d[e] += 1
                except:
                    d[e] = 1
            r = []
            for k, v in d.items():
                if v%2:
                    r.append(k)
            return r

        new = Signal(32)
        self.comb += new.eq(self.last ^ self.data)

        # compute and optimize the parallel implementation of the CRC's LFSR
        taps = [x for x in range(width) if (1 << x) & polynom]
        curval = [[("new", i)] for i in range(width)]
        for i in range(width):
            feedback = curval.pop()
            for j in range(width-1):
                if j + 1 in taps:
                    curval[j] += feedback
                curval[j] = _optimize_xors(curval[j])
            curval.insert(0, feedback)

        # implement logic
        for i in range(width):
            xors = []
            for t, n in curval[i]:
                if t == "new":
                    xors += [new[n]]
            self.comb += self.next[i].eq(reduce(xor, xors))


@ResetInserter()
@CEInserter()
class LiteSATACRC(Module):
    """SATA CRC

    Implement a SATA CRC generator/checker

    Attributes
    ----------
    value : out
        CRC value (used for generator).
    error : out
        CRC error (used for checker).
    """
    width   = 32
    polynom = 0x04C11DB7
    init    = 0x52325032
    check   = 0x00000000

    def __init__(self):
        self.data  = Signal(self.width)
        self.value = Signal(self.width)
        self.error = Signal()

        # # #

        engine = CRCEngine(self.width, self.polynom)
        self.submodules += engine
        reg_i = Signal(self.width, reset=self.init)
        self.sync += reg_i.eq(engine.next)
        self.comb += [
            engine.data.eq(self.data),
            engine.last.eq(reg_i),

            self.value.eq(reg_i),
            self.error.eq(engine.next != self.check)
        ]


class LiteSATACRCInserter(Module):
    """SATA CRC Inserter

    Append a CRC at the end of each packet.

    Parameters
    ----------
    layout : layout
        Layout of the dataflow.

    Attributes
    ----------
    sink : in
        Packets input without CRC.
    source : out
        Packets output with CRC.
    """
    def __init__(self, description):
        self.sink   = sink   = stream.Endpoint(description)
        self.source = source = stream.Endpoint(description)
        self.busy   = Signal()

        # # #

        crc = LiteSATACRC()
        fsm = FSM(reset_state="IDLE")
        self.submodules += crc, fsm

        fsm.act("IDLE",
            crc.reset.eq(1),
            sink.ready.eq(1),
            If(sink.valid,
                sink.ready.eq(0),
                NextState("COPY"),
            )
        )
        fsm.act("COPY",
            crc.ce.eq(sink.valid & source.ready),
            crc.data.eq(sink.data),
            sink.connect(source),
            source.last.eq(0),
            If(sink.valid & sink.last & source.ready,
                NextState("INSERT"),
            )
        )
        fsm.act("INSERT",
            source.valid.eq(1),
            source.last.eq(1),
            source.data.eq(crc.value),
            If(source.ready, NextState("IDLE"))
        )
        self.comb += self.busy.eq(~fsm.ongoing("IDLE"))


class LiteSATACRCChecker(Module):
    """SATA CRC Checker

    Check CRC at the end of each packet.

    Parameters
    ----------
    layout : layout
        Layout of the dataflow.

    Attributes
    ----------
    sink : in
        Packets input with CRC.
    source : out
        Packets output without CRC and "error" set to 0
        on last when CRC OK / set to 1 when CRC KO.
    """
    def __init__(self, description):
        self.sink   = sink   = stream.Endpoint(description)
        self.source = source = stream.Endpoint(description)
        self.busy   = Signal()

        # # #

        crc = LiteSATACRC()
        self.submodules += crc

        error = Signal()
        fifo  = ResetInserter()(stream.SyncFIFO(description, 2))
        self.submodules += fifo

        fsm = FSM(reset_state="RESET")
        self.submodules += fsm

        fifo_in   = Signal()
        fifo_out  = Signal()
        fifo_full = Signal()

        self.comb += [
            fifo_full.eq(fifo.level == 1),
            fifo_in.eq(sink.valid & (~fifo_full | fifo_out)),
            fifo_out.eq(source.valid & source.ready),

            sink.connect(fifo.sink),
            fifo.sink.valid.eq(fifo_in),
            self.sink.ready.eq(fifo_in),

            source.valid.eq(sink.valid & fifo_full),
            source.last.eq(sink.last),
            fifo.source.ready.eq(fifo_out),
            source.payload.eq(fifo.source.payload),

            source.error.eq(sink.error | (crc.error & source.last)),
        ]

        fsm.act("RESET",
            crc.reset.eq(1),
            fifo.reset.eq(1),
            NextState("IDLE"),
        )
        fsm.act("IDLE",
            crc.data.eq(sink.data),
            If(sink.valid & sink.ready,
                crc.ce.eq(1),
                NextState("COPY")
            )
        )
        fsm.act("COPY",
            crc.data.eq(sink.data),
            If(sink.valid & sink.ready,
                crc.ce.eq(1),
                If(sink.last,
                    NextState("RESET")
                )
            )
        )
        self.comb += self.busy.eq(~fsm.ongoing("IDLE"))

# Link Scrambling ----------------------------------------------------------------------------------

@CEInserter()
class Scrambler(Module):
    """SATA Scrambler

    Implement a SATA Scrambler

    Attributes
    ----------
    value : out
        Scrambled value.
    """
    def __init__(self):
        self.value = Signal(32)

        # # #

        context    = Signal(16, reset=0xf0f6)
        next_value = Signal(32)
        self.sync += context.eq(next_value[16:32])

        # XXX: from SATA specification, replace it with
        # a generic implementation using polynoms.
        lfsr_coefs = (
            (15, 13, 4, 0),  # 0
            (15, 14, 13, 5, 4, 1, 0),
            (14, 13, 6, 5, 4, 2, 1, 0),
            (15, 14, 7, 6, 5, 3, 2, 1),
            (13, 8, 7, 6, 3, 2, 0),
            (14, 9, 8, 7, 4, 3, 1),
            (15, 10, 9, 8, 5, 4, 2),
            (15, 13, 11, 10, 9, 6, 5, 4, 3, 0),
            (15, 14, 13, 12, 11, 10, 7, 6, 5, 1, 0),
            (14, 12, 11, 8, 7, 6, 4, 2, 1, 0),
            (15, 13, 12, 9, 8, 7, 5, 3, 2, 1),
            (15, 14, 10, 9, 8, 6, 3, 2, 0),
            (13, 11, 10, 9, 7, 3, 1, 0),
            (14, 12, 11, 10, 8, 4, 2, 1),
            (15, 13, 12, 11, 9, 5, 3, 2),
            (15, 14, 12, 10, 6, 3, 0),

            (11, 7, 1, 0),  # 16
            (12, 8, 2, 1),
            (13, 9, 3, 2),
            (14, 10, 4, 3),
            (15, 11, 5, 4),
            (15, 13, 12, 6, 5, 4, 0),
            (15, 14, 7, 6, 5, 4, 1, 0),
            (13, 8, 7, 6, 5, 4, 2, 1, 0),
            (14, 9, 8, 7, 6, 5, 3, 2, 1),
            (15, 10, 9, 8, 7, 6, 4, 3, 2),
            (15, 13, 11, 10, 9, 8, 7, 5, 3, 0),
            (15, 14, 13, 12, 11, 10, 9, 8, 6, 1, 0),
            (14, 12, 11, 10, 9, 7, 4, 2, 1, 0),
            (15, 13, 12, 11, 10, 8, 5, 3, 2, 1),
            (15, 14, 12, 11, 9, 6, 3, 2, 0),
            (12, 10, 7, 3, 1, 0),
        )

        for n, coefs in enumerate(lfsr_coefs):
            eq = [context[i] for i in coefs]
            self.comb += next_value[n].eq(reduce(xor, eq))

        self.comb += self.value.eq(next_value)


@ResetInserter()
class LiteSATAScrambler(Module):
    def __init__(self, description):
        self.sink   = sink   = stream.Endpoint(description)
        self.source = source = stream.Endpoint(description)

        # # #

        scrambler = Scrambler()
        self.submodules += scrambler
        self.comb += [
            scrambler.ce.eq(sink.valid & sink.ready),
            sink.connect(source),
            source.data.eq(sink.data ^ scrambler.value)
        ]

# Link Clock Compensation --------------------------------------------------------------------------

class LiteSATACONTInserter(Module):
    def __init__(self, description):
        self.sink   = sink   = stream.Endpoint(description)
        self.source = source = stream.Endpoint(description)

        # # #

        counter       = Signal(max=4)
        counter_reset = Signal()
        counter_ce    = Signal()
        self.sync += \
            If(counter_reset,
                counter.eq(0)
            ).Elif(counter_ce,
                counter.eq(counter + 1)
            )

        is_data  = Signal()
        was_data = Signal()
        was_hold = Signal()
        change   = Signal()
        self.comb += is_data.eq(sink.charisk == 0)

        last_data = Signal(32)
        last_primitive = Signal(32)
        last_charisk = Signal(4)
        self.sync += [
            If(sink.valid & source.ready,
                last_data.eq(sink.data),
                last_charisk.eq(sink.charisk),
                If(~is_data,
                    last_primitive.eq(sink.data),
                ),
                was_data.eq(is_data),
                was_hold.eq(last_primitive == primitives["HOLD"])
            )
        ]
        self.comb += change.eq(
            (sink.data != last_data) |
            (sink.charisk != last_charisk) |
            is_data
        )

        # Scrambler
        scrambler = ResetInserter()(Scrambler())
        self.submodules += scrambler

        # Datapath
        self.comb += [
            sink.connect(source),
            If(sink.valid,
                If(~change,
                    counter_ce.eq(sink.ready & (counter != 2)),
                    # insert CONT
                    If(counter == 1,
                        source.charisk.eq(0b0001),
                        source.data.eq(primitives["CONT"])
                    # insert scrambled data for EMI
                    ).Elif(counter == 2,
                        scrambler.ce.eq(sink.ready),
                        source.charisk.eq(0b0000),
                        source.data.eq(scrambler.value)
                    )
                ).Else(
                    counter_reset.eq(source.ready),
                    If(counter == 2,
                        # Reinsert last primitive
                        If(is_data | (~is_data & was_hold),
                            source.valid.eq(1),
                            sink.ready.eq(0),
                            source.charisk.eq(0b0001),
                            source.data.eq(last_primitive)
                        )
                    )
                )
            )
        ]

class LiteSATACONTRemover(Module):
    def __init__(self, description):
        self.sink   = sink   = stream.Endpoint(description)
        self.source = source = stream.Endpoint(description)

        # # #

        is_data      = Signal()
        is_cont      = Signal()
        in_cont      = Signal()
        cont_ongoing = Signal()

        self.comb += [
            is_data.eq(sink.charisk == 0),
            is_cont.eq(~is_data & (sink.data == primitives["CONT"]))
        ]
        self.sync += \
            If(sink.valid & sink.ready,
                If(is_cont,
                    in_cont.eq(1)
                ).Elif(~is_data,
                    in_cont.eq(0)
                )
            )
        self.comb += cont_ongoing.eq(is_cont | (in_cont & is_data))

        # Datapath
        last_primitive = Signal(32)
        self.sync += [
            If(sink.valid & sink.ready,
                If(~is_data & ~is_cont,
                    last_primitive.eq(sink.data)
                )
            )
        ]
        self.comb += [
            sink.connect(source),
            If(cont_ongoing,
                source.charisk.eq(0b0001),
                source.data.eq(last_primitive)
            )
        ]

# Link Alignment -----------------------------------------------------------------------------------

class LiteSATAALIGNInserter(Module):
    def __init__(self, description):
        self.sink   = sink   = stream.Endpoint(description)
        self.source = source = stream.Endpoint(description)

        # # #

        # send 2 ALIGN every 256 DWORDs
        # used for clock compensation between
        # HOST and device
        cnt  = Signal(8)
        send = Signal()
        self.sync += \
            If(source.valid & source.ready,
                cnt.eq(cnt+1)
            )
        self.comb += [
            send.eq(cnt < 2),
            If(send,
                source.valid.eq(1),
                source.charisk.eq(0b0001),
                source.data.eq(primitives["ALIGN"]),
                sink.ready.eq(0)
            ).Else(
                source.valid.eq(sink.valid),
                source.data.eq(sink.data),
                source.charisk.eq(sink.charisk),
                sink.ready.eq(source.ready)
            )
        ]


class LiteSATAALIGNRemover(Module):
    def __init__(self, description):
        self.sink   = sink   = stream.Endpoint(description)
        self.source = source = stream.Endpoint(description)

        # # #

        charisk_match = sink.charisk == 0b0001
        data_match    = sink.data == primitives["ALIGN"]

        self.comb += \
            If(sink.valid & charisk_match & data_match,
                sink.ready.eq(1),
            ).Else(
                sink.connect(source)
            )

# Link TX ------------------------------------------------------------------------------------------

from_rx = [
    ("idle",            1),
    ("insert",          32),
    ("primitive_valid", 1),
    ("primitive",       32)
]

class LiteSATALinkTX(Module):
    def __init__(self):
        self.sink    = sink   = stream.Endpoint(link_description(32))
        self.source  = source = stream.Endpoint(phy_description(32))
        self.from_rx = stream.Endpoint(from_rx)

        self.error = Signal()

        # # #

        # CRC / Scrambler
        crc       = LiteSATACRCInserter(link_description(32))
        scrambler = LiteSATAScrambler(link_description(32))
        pipeline  = Pipeline(sink, crc, scrambler)
        self.submodules += crc, scrambler, pipeline

        # datas / primitives mux
        insert = Signal(32)
        copy   = Signal()
        self.comb += [
            If(self.from_rx.insert,
                source.valid.eq(1),
                source.data.eq(self.from_rx.insert),
                source.charisk.eq(0b0001),
            ).Elif(insert,
                source.valid.eq(1),
                source.data.eq(insert),
                source.charisk.eq(0b0001),
            ).Elif(copy,
                source.valid.eq(1),
                pipeline.source.ready.eq(source.ready),
                If(pipeline.source.valid,
                    source.data.eq(pipeline.source.data),
                    source.charisk.eq(0b0000)
                ).Else(
                    source.data.eq(primitives["HOLD"]),
                    source.charisk.eq(0b0001)
                )
            )
        ]

        # FSM
        self.submodules.fsm = fsm = FSM(reset_state="IDLE")
        fsm.act("IDLE",
            scrambler.reset.eq(1),
            If(self.from_rx.idle,
                insert.eq(primitives["SYNC"]),
                If(pipeline.source.valid,
                    If(self.from_rx.primitive_valid &
                       (self.from_rx.primitive == primitives["SYNC"]),
                        NextState("RDY")
                    )
                )
            )
        )
        fsm.act("RDY",
            insert.eq(primitives["X_RDY"]),
            If(~self.from_rx.idle,
                NextState("IDLE")
            ).Elif(self.from_rx.primitive_valid &
                   (self.from_rx.primitive == primitives["R_RDY"]),
                NextState("SOF")
            )
        )
        fsm.act("SOF",
            insert.eq(primitives["SOF"]),
            If(source.ready,
                NextState("COPY")
            )
        )
        fsm.act("COPY",
            copy.eq(1),
            If(pipeline.source.valid &
               pipeline.source.last &
               pipeline.source.ready,
                NextState("EOF")
            ).Elif(self.from_rx.primitive_valid &
               (self.from_rx.primitive == primitives["HOLD"]),
               NextState("HOLDA")
            )
        )
        fsm.act("HOLDA",
            insert.eq(primitives["HOLDA"]),
            If(self.from_rx.primitive_valid &
               (self.from_rx.primitive != primitives["HOLD"]),
                NextState("COPY")
            ).Elif(self.error,
                NextState("IDLE")
            )
        )
        fsm.act("EOF",
            insert.eq(primitives["EOF"]),
            If(source.ready,
                NextState("WTRM")
            )
        )
        fsm.act("WTRM",
            insert.eq(primitives["WTRM"]),
            If(self.from_rx.primitive_valid,
                If(self.from_rx.primitive == primitives["R_OK"],
                    NextState("IDLE")
                ).Elif(self.from_rx.primitive == primitives["R_ERR"],
                    NextState("IDLE")
                )
            ).Elif(self.error,
                NextState("IDLE")
            )
        )

        # error detection
        self.sync += [
            # generate error if receiving SYNC during transfer (disk returns to IDLE)
            If(~(fsm.ongoing("IDLE") | fsm.ongoing("RDY")),
                self.error.eq(self.from_rx.primitive_valid &
                              (self.from_rx.primitive == primitives["SYNC"]))
            )
        ]

# Link RX ------------------------------------------------------------------------------------------

class LiteSATALinkRX(Module):
    def __init__(self):
        self.sink   = sink   = stream.Endpoint(phy_description(32))
        self.source = source = stream.Endpoint(link_description(32))
        self.hold   = Signal()
        self.to_tx  = stream.Endpoint(from_rx)

        # # #

        # Always ready from phy
        self.comb += sink.ready.eq(1)

        # Datas / primitives detection
        insert          = Signal(32)
        data_valid      = Signal()
        primitive_valid = Signal()
        primitive       = Signal(32)
        self.comb += [
            If(sink.valid,
                data_valid.eq(sink.charisk == 0),
                primitive_valid.eq(sink.charisk == 0b0001)
            ),
            primitive.eq(sink.data)
        ]

        # Descrambler / CRC
        descrambler = LiteSATAScrambler(link_description(32))
        crc         = LiteSATACRCChecker(link_description(32))
        pipeline    = Pipeline(descrambler, crc, source)
        self.submodules += descrambler, crc, pipeline

        # Internal logic
        self.crc_error = crc_error = Signal()
        self.sync += \
            If(crc.source.valid & crc.source.last & crc.source.ready,
                crc_error.eq(crc.source.error)
            )

        # FSM
        self.submodules.fsm = fsm = FSM(reset_state="IDLE")
        fsm.act("IDLE",
            descrambler.reset.eq(1),
            If(primitive_valid &
               (primitive == primitives["X_RDY"]),
                NextState("RDY")
            )
        )
        fsm.act("RDY",
            insert.eq(primitives["R_RDY"]),
            If(primitive_valid &
               (primitive == primitives["SOF"]),
                NextState("WAIT_FIRST")
            )
        )
        fsm.act("WAIT_FIRST",
            insert.eq(primitives["R_IP"]),
            If(data_valid,
                NextState("COPY")
            )
        )
        fsm.act("COPY",
            pipeline.sink.valid.eq(data_valid),
            insert.eq(primitives["R_IP"]),
            If(primitive_valid & (primitive != primitives["HOLDA"]),
                If(primitive == primitives["HOLD"],
                    insert.eq(primitives["HOLDA"])
                ).Elif(primitive == primitives["EOF"],
                    # 1 clock cycle latency
                    pipeline.sink.valid.eq(1),
                    pipeline.sink.last.eq(1),
                    NextState("WTRM")
                )
            ).Elif(self.hold,
                insert.eq(primitives["HOLD"])
            )
        )
        # 1 clock cycle latency
        self.sync += If(data_valid, pipeline.sink.data.eq(sink.data))
        fsm.act("EOF",
            insert.eq(primitives["R_IP"]),
            If(primitive_valid &
               (primitive == primitives["WTRM"]),
                NextState("WTRM")
            )
        )
        fsm.act("WTRM",
            insert.eq(primitives["R_IP"]),
            If(~crc_error,
                NextState("R_OK")
            ).Else(
                NextState("R_ERR")
            )
        )
        fsm.act("R_OK",
            insert.eq(primitives["R_OK"]),
            If(primitive_valid &
               (primitive == primitives["SYNC"]),
                NextState("IDLE")
            )
        )
        fsm.act("R_ERR",
            insert.eq(primitives["R_ERR"]),
            If(primitive_valid &
               (primitive == primitives["SYNC"]),
                NextState("IDLE")
            )
        )

        # To TX
        self.comb += [
            self.to_tx.idle.eq(fsm.ongoing("IDLE")),
            self.to_tx.insert.eq(insert),
            self.to_tx.primitive_valid.eq(primitive_valid),
            self.to_tx.primitive.eq(primitive)
        ]

# Link ---------------------------------------------------------------------------------------------

class LiteSATALink(Module):
    def __init__(self, phy):
        # TX ---------------------------------------------------------------------------------------
        self.submodules.tx = BufferizeEndpoints({"source": DIR_SOURCE})(LiteSATALinkTX())
        self.submodules.tx_align = LiteSATAALIGNInserter(phy_description(32))
        self.submodules.tx_pipeline = Pipeline(self.tx, self.tx_align, phy)

        # RX ---------------------------------------------------------------------------------------
        self.submodules.rx_align = LiteSATAALIGNRemover(phy_description(32))
        self.submodules.rx_cont = LiteSATACONTRemover(phy_description(32))
        self.submodules.rx = BufferizeEndpoints({"sink": DIR_SINK})(LiteSATALinkRX())
        self.submodules.rx_buffer = stream.SyncFIFO(link_description(32), 256)
        self.submodules.rx_pipeline = Pipeline(phy, self.rx_align, self.rx_cont, self.rx, self.rx_buffer)

        # RX --> TX --------------------------------------------------------------------------------
        self.comb += self.rx.to_tx.connect(self.tx.from_rx)

        self.sink, self.source = self.tx_pipeline.sink, self.rx_pipeline.source

        # Hold -------------------------------------------------------------------------------------
        self.comb += self.rx.hold.eq(self.rx_buffer.level > 128)



#
# This file is part of LiteSATA.
#
# Copyright (c) 2015-2016 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

import unittest
import subprocess

from litesata.common import *
from litesata.core.link import Scrambler

from litex.soc.interconnect.stream_sim import *


class TestLinkScrambler(unittest.TestCase):
    def test_link_scrambler(self):
        def generator(dut):
            # Init CRC
            yield dut.scrambler.ce.eq(1)
            yield dut.scrambler.reset.eq(1)
            yield
            yield dut.scrambler.reset.eq(0)

            # Log results
            yield
            sim_values = []
            for i in range(dut.length):
                sim_values.append((yield dut.scrambler.value))
                yield

            # Stop
            yield dut.scrambler.ce.eq(0)
            for i in range(32):
                yield

            # Get C code reference
            c_values = dut.get_c_values(dut.length)

            # Check results
            s, l, e = check(c_values, sim_values)
            print("shift " + str(s) + " / length " + str(l) + " / errors " + str(e))
            self.assertEqual(s, 0)
            self.assertEqual(e, 0)

        class DUT(Module):
            def __init__(self, length):
                self.submodules.scrambler = ResetInserter()(Scrambler())
                self.length = length

            def get_c_values(self, length):
                stdin = "0x{:08x}".format(length)
                with subprocess.Popen("./test/model/scrambler",
                    stdin  = subprocess.PIPE,
                    stdout = subprocess.PIPE) as process:
                    process.stdin.write(stdin.encode("ASCII"))
                    out, err = process.communicate()
                return [int(e, 16) for e in out.decode("ASCII").split("\n")[:-1]]

        dut = DUT(1024)
        generators = {
            "sys" :   [generator(dut)]
        }
        clocks = {"sys": 10}
        run_simulation(dut, generators, clocks)



#
# This file is part of LiteSATA.
#
# Copyright (c) 2015-2019 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

import unittest

import subprocess

from litesata.common import *
from litesata.core.link import LiteSATACRC

from litex.soc.interconnect.stream_sim import *

class TestLinkCRC(unittest.TestCase):
    def test_link_crc(self):
        def generator(dut):
            dut.errors = 0
            # Init CRC
            yield dut.crc.data.eq(0)
            yield dut.crc.ce.eq(1)
            yield dut.crc.reset.eq(1)
            yield
            yield dut.crc.reset.eq(0)

            # Feed CRC with datas
            datas = []
            for i in range(dut.length):
                data = seed_to_data(i, dut.random)
                datas.append(data)
                yield dut.crc.data.eq(data)
                yield

            # Log results
            yield
            sim_crc = (yield dut.crc.value)

            # Stop
            yield dut.crc.ce.eq(0)
            for i in range(32):
                yield

            # Get C code reference
            c_crc = dut.get_c_crc(datas)

            # Check results
            s, l, e = check(c_crc, sim_crc)
            print("shift " + str(s) + " / length " + str(l) + " / errors " + str(e))
            self.assertEqual(s, 0)
            self.assertEqual(e, 0)

        class DUT(Module):
            def __init__(self, length, random):
                self.submodules.crc = LiteSATACRC()
                self.length = length
                self.random = random

            def get_c_crc(self, datas):
                stdin = ""
                for data in datas:
                    stdin += "0x{:08x} ".format(data)
                stdin += "exit"
                with subprocess.Popen("./test/model/crc", stdin=subprocess.PIPE, stdout=subprocess.PIPE) as process:
                    process.stdin.write(stdin.encode("ASCII"))
                    out, err = process.communicate()
                return int(out.decode("ASCII"), 16)

        dut = DUT(1024, False)
        run_simulation(dut, generator(dut))



#
# This file is part of LiteSATA.
#
# Copyright (c) 2015-2020 Florent Kermarrec <florent@enjoy-digital.fr>
# Copyright (c) 2016 Olof Kindgren <olof.kindgren@gmail.com>
# SPDX-License-Identifier: BSD-2-Clause

from litesata.common import *
from litesata.core.link import Scrambler
from litesata.frontend.identify import LiteSATAIdentify, LiteSATAIdentifyCSR

from litex.soc.interconnect.csr import *

# LiteSATABISTGenerator ----------------------------------------------------------------------------

class LiteSATABISTGenerator(Module):
    def __init__(self, user_port, count_width=32):
        self.start   = Signal()
        self.sector  = Signal(48)
        self.count   = Signal(16)
        self.random  = Signal()

        self.done    = Signal()
        self.aborted = Signal()
        self.errors  = Signal(32)  # Note: Not used for writes

        # # #

        n          = user_port.dw//32
        count_mult = user_port.dw//user_port.controller_dw
        if count_mult == 0:
            raise ValueError

        source, sink = user_port.sink, user_port.source

        count     = Signal(count_width)
        scrambler = ResetInserter()(Scrambler())
        self.submodules += scrambler

        self.submodules.fsm = fsm = FSM(reset_state="IDLE")
        fsm.act("IDLE",
            self.done.eq(1),
            scrambler.reset.eq(1),
            NextValue(count, 0),
            If(self.start,
                NextState("SEND-CMD-AND-DATA")
            )
        )
        self.comb += [
            source.last.eq(count == (logical_sector_size//4*self.count)-1),
            source.write.eq(1),
            source.sector.eq(self.sector),
            source.count.eq(self.count*count_mult),
            If(self.random,
                source.data.eq(Replicate(scrambler.value, n))
            ).Else(
                source.data.eq(Replicate(count, n))
            )
        ]
        fsm.act("SEND-CMD-AND-DATA",
            source.valid.eq(1),
            If(source.valid & source.ready,
                scrambler.ce.eq(1),
                NextValue(count, count + 1),
                If(source.last,
                    NextState("WAIT-ACK")
                )
            )
        )
        fsm.act("WAIT-ACK",
            sink.ready.eq(1),
            If(sink.valid,
                NextState("IDLE")
            )
        )

        self.sync += [
            If(self.start,
                self.aborted.eq(0)
            ).Elif(sink.valid & sink.ready,
                self.aborted.eq(self.aborted | sink.failed)
            )
        ]

# LiteSATABISTChecker ------------------------------------------------------------------------------

class LiteSATABISTChecker(Module):
    def __init__(self, user_port, count_width=32):
        self.start   = Signal()
        self.sector  = Signal(48)
        self.count   = Signal(16)
        self.random  = Signal()

        self.done    = Signal()
        self.aborted = Signal()
        self.errors  = Signal(count_width)

        # # #

        n          = user_port.dw//32
        count_mult = user_port.dw//user_port.controller_dw
        if count_mult == 0:
            raise ValueError

        source, sink = user_port.sink, user_port.source

        count  = Signal(count_width)
        errors = Signal(count_width)
        self.comb += self.errors.eq(errors)

        scrambler = ResetInserter()(Scrambler())
        self.submodules += scrambler

        self.submodules.fsm = fsm = FSM(reset_state="IDLE")
        fsm.act("IDLE",
            self.done.eq(1),
            scrambler.reset.eq(1),
            NextValue(count, 0),
            If(self.start,
                NextValue(errors, 0),
                NextState("SEND-CMD")
            )
        )
        self.comb += [
            source.last.eq(1),
            source.read.eq(1),
            source.sector.eq(self.sector),
            source.count.eq(self.count*count_mult),
        ]
        fsm.act("SEND-CMD",
            source.valid.eq(1),
            If(source.ready,
                NextState("WAIT-ACK")
            )
        )
        fsm.act("WAIT-ACK",
            If(sink.valid & sink.read,
                NextState("RECEIVE-DATA")
            )
        )
        expected_data = Signal(n*32)
        self.comb += [
            If(self.random,
                expected_data.eq(Replicate(scrambler.value, n))
            ).Else(
                expected_data.eq(Replicate(count, n))
            )
        ]
        fsm.act("RECEIVE-DATA",
            sink.ready.eq(1),
            If(sink.valid,
                scrambler.ce.eq(1),
                NextValue(count, count + 1),
                If(sink.data != expected_data,
                    If(~sink.end,
                        NextValue(errors, errors + 1)
                    )
                ),
                If(sink.end,
                    If(sink.last,
                        NextState("IDLE")
                    ).Else(
                        NextState("WAIT-ACK")
                    )
                )
            )
        )

        self.sync += [
            If(self.start,
                self.aborted.eq(0)
            ).Elif(sink.valid & sink.ready,
                self.aborted.eq(self.aborted | sink.failed)
            )
        ]

# LiteSATABISTUnitCSR ------------------------------------------------------------------------------

class LiteSATABISTUnitCSR(Module, AutoCSR):
    def __init__(self, bist_unit):
        self._start   = CSR()
        self._sector  = CSRStorage(48)
        self._count   = CSRStorage(16)
        self._loops   = CSRStorage(8)
        self._random  = CSRStorage()

        self._done    = CSRStatus()
        self._aborted = CSRStatus()
        self._errors  = CSRStatus(32)
        self._cycles  = CSRStatus(32)

        # # #

        self.submodules += bist_unit

        start = self._start.r & self._start.re
        done  = self._done.status
        loops = self._loops.storage

        self.comb += [
            bist_unit.sector.eq(self._sector.storage),
            bist_unit.count.eq(self._count.storage),
            bist_unit.random.eq(self._random.storage),

            self._aborted.status.eq(bist_unit.aborted),
            self._errors.status.eq(bist_unit.errors)
        ]

        loop   = Signal(8)
        cycles = Signal(32)
        self.comb += self._cycles.status.eq(cycles)

        self.submodules.fsm = fsm = FSM(reset_state="IDLE")
        fsm.act("IDLE",
            self._done.status.eq(1),
            NextValue(loop, 0),
            If(start,
                NextValue(cycles, 0),
                NextState("CHECK")
            )
        )
        fsm.act("CHECK",
            If(loop < loops,
                NextState("START")
            ).Else(
                NextState("IDLE")
            ),
            NextValue(cycles, cycles + 1),
        )
        fsm.act("START",
            bist_unit.start.eq(1),
            NextState("WAIT-DONE"),
            NextValue(cycles, cycles + 1),
        )
        fsm.act("WAIT-DONE",
            If(bist_unit.done,
                NextValue(loop, loop + 1),
                NextState("CHECK")
            ),
            NextValue(cycles, cycles + 1),
        )

# LiteSATABIST --------------------------------------------------------------------------

class LiteSATABIST(Module, AutoCSR):
    def __init__(self, crossbar, with_csr=False, count_width=32):
        generator = LiteSATABISTGenerator(crossbar.get_port(), count_width)
        checker   = LiteSATABISTChecker(crossbar.get_port(), count_width)
        identify  = LiteSATAIdentify(crossbar.get_port())
        if with_csr:
            generator = LiteSATABISTUnitCSR(generator)
            checker   = LiteSATABISTUnitCSR(checker)
            identify  = LiteSATAIdentifyCSR(identify)
        self.submodules.generator = generator
        self.submodules.checker   = checker
        self.submodules.identify  = identify



#
# This file is part of LiteSATA.
#
# Copyright (c) 2015-2019 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

from litesata.common import *

from litex.gen.genlib.misc import WaitTimer

# LiteSATAPHYDatapathRX ----------------------------------------------------------------------------

class LiteSATAPHYDatapathRX(Module):
    """SATA PHY RX datapath

    Manages the RX datapath between the transceiver and the SATA core.

    This modules receives data/special characters from the transceiver and needs
    to:
    - realign received data on 32 bits word boundaries.
    - change clock domain from "sata_rx" to "sys".

    It handles both 32 bits and 16 bits transceiver data width. The realignment is
    done by storing two complete words from the transceiver, detecting the byte
    alignement with the special character and use this information to select the
    data we will output.
    """
    def __init__(self, data_width):
        self.sink   = sink   = stream.Endpoint(phy_description(data_width))
        self.source = source = stream.Endpoint(phy_description(32))

        # # #

        # width convertion and byte alignment
        byte_alignment = Signal(data_width//8)
        last_charisk   = Signal(data_width//8)
        last_data      = Signal(data_width)
        sr_charisk     = Signal(2*data_width//8)
        sr_data        = Signal(2*data_width)
        self.sync.sata_rx += \
            If(sink.valid & sink.ready,
                If(sink.charisk != 0,
                    byte_alignment.eq(sink.charisk)
                ),
                last_charisk.eq(sink.charisk),
                last_data.eq(sink.data)
            )
        self.comb += [
            sr_charisk.eq(Cat(last_charisk, sink.charisk)),
            sr_data.eq(Cat(last_data, sink.data)),
        ]


        converter = StrideConverter(
            description_from = phy_description(data_width),
            description_to   = phy_description(32),
            reverse          = False,
        )
        if data_width == 16: # when data_width=32, converter is just direct connection
            converter = ResetInserter()(ClockDomainsRenamer("sata_rx")(converter))
        self.submodules += converter
        cases = {}
        for i in range(data_width//8):
            cases[2**i] = [
                converter.sink.charisk.eq(sr_charisk[data_width//8-i:2*data_width//8-i]),
                converter.sink.data.eq(sr_data[data_width-8*i:2*data_width-8*i])
            ]
        self.comb += [
            converter.sink.valid.eq(sink.valid),
            Case(byte_alignment, cases),
            sink.ready.eq(converter.sink.ready)
        ]
        if data_width == 16:
            self.comb += converter.reset.eq(converter.source.charisk[2:] != 0)

        # clock domain crossing
        #   (gen3) 300MHz (16 bits) / 150MHz (32 bits) sata_rx clk to sys_clk
        #   (gen2) 150MHz (16 bits) / 75MHz (32 bits) sata_rx clk to sys_clk
        #   (gen1) 75MHz (16 bits) / 37.5MHz (32 bits) sata_rx clk to sys_clk
        #   requirements:
        #     due to the convertion ratio of 2, sys_clk need to be > sata_rx/2
        #     source destination is always able to accept data (ready always 1)
        fifo = stream.ClockDomainCrossing(
            layout  = phy_description(32),
            cd_from = "sata_rx",
            cd_to   = "sys",
            depth   = 8,
            with_common_rst = True,
        )
        self.submodules += fifo
        self.comb += [
            converter.source.connect(fifo.sink),
            fifo.source.connect(source)
        ]

# LiteSATAPHYDatapathTX ----------------------------------------------------------------------------

class LiteSATAPHYDatapathTX(Module):
    """SATA PHY TX Datapath

    Manages the TX Datapath between SATA core and the transceiver.

    This modules receives datas / special characters from the core and needs
    to:
    - convert 32 bits data to the transceiver data width (32 bits or 16 bits)
    - change clock domain from "sys" to "sata_rx".
    """
    def __init__(self, data_width):
        self.sink   = sink   = stream.Endpoint(phy_description(32))
        self.source = source = stream.Endpoint(phy_description(data_width))

        # # #

        # clock domain crossing
        #   (gen3) sys_clk to 300MHz (16 bits) / 150MHz (32 bits) sata_tx clk
        #   (gen2) sys_clk to 150MHz (16 bits) / 75MHz (32 bits) sata_tx clk
        #   (gen1) sys_clk to 75MHz (16 bits) / 37.5MHz (32 bits) sata_tx clk
        #   requirements:
        #     source destination is always able to accept data (ready always 1)
        fifo = stream.ClockDomainCrossing(
            layout  = phy_description(32),
            cd_from = "sys",
            cd_to   = "sata_tx",
            depth   = 8,
            with_common_rst = True,
        )
        self.submodules += fifo
        self.comb += sink.connect(fifo.sink)

        # width convertion
        converter = StrideConverter(
            description_from = phy_description(32),
            description_to   = phy_description(data_width),
            reverse          = False,
        )
        converter = ClockDomainsRenamer("sata_tx")(converter)
        self.submodules += converter
        self.comb += [
            fifo.source.connect(converter.sink),
            converter.source.connect(source)
        ]

# LiteSATAPHYAlignTimer ----------------------------------------------------------------------------

class LiteSATAPHYAlignTimer(Module):
    """SATA PHY align timer

    This modules detects ALIGN primitives that we receive from the device and
    decide whether or not our device is returning valid data (Using RX idle
    signal from the transceiver is not recommended by vendors).
    """
    def __init__(self):
        self.sink = sink = stream.Endpoint(phy_description(32))

        # # #

        self.submodules.timer = WaitTimer(256*16)

        charisk_match = sink.charisk == 0b0001
        data_match    = sink.data == primitives["ALIGN"]

        self.comb += \
            If(sink.valid &
              (sink.charisk == 0b0001) &
              (sink.data == primitives["ALIGN"]),
                self.timer.wait.eq(0)
            ).Else(
                self.timer.wait.eq(1),
            )

# LiteSATAPHYDatapath ------------------------------------------------------------------------------

class LiteSATAPHYDatapath(Module):
    """SATA PHY datapath

    Manages the datapath between the core and the transceiver.

    This modules does mainly data width convertion (core is always 32 bits,
    transceiver can either be 16 bits or 32 bits depending on the FPGA speedgrade)
    and the cross domain crossing:
    - "sys" to "sata_tx" for the TX datapath.
    - "sata_rx" to "sys" for the RX datapath.

    Misalign and Idle signals are also generated to have a state of the link and
    are used by the SATA PHY controller

    """
    def __init__(self, trx, ctrl):
        self.sink   = sink   = stream.Endpoint(phy_description(32))
        self.source = source = stream.Endpoint(phy_description(32))

        self.misalign = Signal()
        self.rx_idle  = Signal()

        # # #

        # TX path
        mux = Multiplexer(phy_description(32), 2)
        tx  = LiteSATAPHYDatapathTX(trx.data_width)
        self.submodules += mux, tx
        self.comb += [
            mux.sel.eq(ctrl.ready),
            ctrl.source.connect(mux.sink0),
            sink.connect(mux.sink1),
            mux.source.connect(tx.sink),
            tx.source.connect(trx.sink)
        ]

        # RX path
        rx    = LiteSATAPHYDatapathRX(trx.data_width)
        demux = Demultiplexer(phy_description(32), 2)
        align_timer = LiteSATAPHYAlignTimer()
        self.submodules += rx, demux, align_timer
        self.comb += [
            demux.sel.eq(ctrl.ready),
            trx.source.connect(rx.sink),
            rx.source.connect(demux.sink),
            rx.source.connect(align_timer.sink, omit=set(["ready"])),
            demux.source0.connect(ctrl.sink),
            demux.source1.connect(source),
        ]

        self.comb += [
            self.misalign.eq(rx.source.valid & ((rx.source.charisk & 0b1110) != 0)),
            self.rx_idle.eq(align_timer.timer.done)
        ]



#
# This file is part of LiteSATA.
#
# Copyright (c) 2015-2019 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

from math import ceil

from migen import *
from migen.genlib.cdc import PulseSynchronizer

from litex.soc.interconnect import stream
from litex.soc.interconnect.stream import *
from litex.soc.interconnect.packet import Header, HeaderField

# Bitrates/Frequencies -----------------------------------------------------------------------------

bitrates = {
    "gen3": 6.0,
    "gen2": 3.0,
    "gen1": 1.5,
}

frequencies = {
    "gen3": 150.0,
    "gen2": 75.0,
    "gen1": 37.5,
}


# PHY / Link Layers --------------------------------------------------------------------------------

primitives = {
    "ALIGN":   0x7b4a4abc,
    "ALIGN_N": 0x7bb5b5bc,
    "CONT":    0x9999aa7c,
    "SYNC":    0xb5b5957c,
    "R_RDY":   0x4a4a957c,
    "R_OK":    0x3535b57c,
    "R_ERR":   0x5656b57c,
    "R_IP":    0x5555b57c,
    "X_RDY":   0x5757b57c,
    "WTRM":    0x5858b57c,
    "SOF":     0x3737b57c,
    "EOF":     0xd5d5b57c,
    "HOLD":    0xd5d5aa7c,
    "HOLDA":   0x9595aa7c
}

def is_primitive(dword):
    for k, v in primitives.items():
        if dword == v:
            return True
    return False


def decode_primitive(dword):
    for k, v in primitives.items():
        if dword == v:
            return k
    return ""


def phy_description(dw):
    layout = [
        ("data", dw),
        ("charisk", dw//8),
    ]
    return EndpointDescription(layout)


def link_description(dw):
    layout = [
        ("data", dw),
        ("error", 1)
    ]
    return EndpointDescription(layout)

class _PulseSynchronizer(PulseSynchronizer):
    def __init__(self, i, idomain, o, odomain):
        PulseSynchronizer.__init__(self, idomain, odomain)
        self.comb += [
            self.i.eq(i),
            o.eq(self.o)
        ]

class _RisingEdge(Module):
    def __init__(self, i, o):
        i_d = Signal()
        self.sync += i_d.eq(i)
        self.comb += o.eq(i & ~i_d)

# Transport Layer ----------------------------------------------------------------------------------

fis_max_dwords = 2048

fis_types = {
    "REG_H2D":          0x27,
    "REG_D2H":          0x34,
    "DMA_ACTIVATE_D2H": 0x39,
    "PIO_SETUP_D2H":    0x5f,
    "DATA":             0x46
}

fis_reg_h2d_header_length = 5
fis_reg_h2d_header_fields = {
    "type":         HeaderField(0*4,  0, 8),
    "pm_port":      HeaderField(0*4,  8, 4),
    "c":            HeaderField(0*4, 15, 1),
    "command":      HeaderField(0*4, 16, 8),
    "features_lsb": HeaderField(0*4, 24, 8),

    "lba_lsb":      HeaderField(1*4, 0, 24),
    "device":       HeaderField(1*4, 24, 8),

    "lba_msb":      HeaderField(2*4, 0, 24),
    "features_msb": HeaderField(2*4, 24, 8),

    "count":        HeaderField(3*4, 0, 16),
    "icc":          HeaderField(3*4, 16, 8),
    "control":      HeaderField(3*4, 24, 8)
}
fis_reg_h2d_header = Header(fis_reg_h2d_header_fields,
                            fis_reg_h2d_header_length,
                            swap_field_bytes=False)

fis_reg_d2h_header_length = 5
fis_reg_d2h_header_fields = {
    "type":    HeaderField(0*4,  0, 8),
    "pm_port": HeaderField(0*4,  8, 4),
    "i":       HeaderField(0*4, 14, 1),
    "status":  HeaderField(0*4, 16, 8),
    "errors":  HeaderField(0*4, 24, 8),

    "lba_lsb": HeaderField(1*4, 0, 24),
    "device":  HeaderField(1*4, 24, 8),

    "lba_msb": HeaderField(2*4, 0, 24),

    "count":   HeaderField(3*4, 0, 16)
}
fis_reg_d2h_header = Header(fis_reg_d2h_header_fields,
                            fis_reg_d2h_header_length,
                            swap_field_bytes=False)

fis_dma_activate_d2h_header_length = 1
fis_dma_activate_d2h_header_fields = {
    "type":    HeaderField(0*4,  0, 8),
    "pm_port": HeaderField(0*4,  8, 4)
}
fis_dma_activate_d2h_header = Header(fis_dma_activate_d2h_header_fields,
                                     fis_dma_activate_d2h_header_length,
                                     swap_field_bytes=False)

fis_pio_setup_d2h_header_length = 5
fis_pio_setup_d2h_header_fields = {
    "type":           HeaderField(0*4,  0, 8),
    "pm_port":        HeaderField(0*4,  8, 4),
    "d":              HeaderField(0*4, 13, 1),
    "i":              HeaderField(0*4, 14, 1),
    "status":         HeaderField(0*4, 16, 8),
    "errors":         HeaderField(0*4, 24, 8),

    "lba_lsb":        HeaderField(1*4, 0, 24),

    "lba_msb":        HeaderField(2*4, 0, 24),

    "count":          HeaderField(3*4, 0, 16),

    "transfer_count": HeaderField(4*4, 0, 16),
}
fis_pio_setup_d2h_header = Header(fis_pio_setup_d2h_header_fields,
                                  fis_pio_setup_d2h_header_length,
                                  swap_field_bytes=False)

fis_data_header_length = 1
fis_data_header_fields = {
    "type": HeaderField(0,  0, 8)
}
fis_data_header = Header(fis_data_header_fields,
                         fis_data_header_length,
                         swap_field_bytes=False)

def transport_tx_description(dw):
    param_layout = [
        ("type",      8),
        ("pm_port",   4),
        ("c",         1),
        ("command",   8),
        ("features", 16),
        ("lba",      48),
        ("device",    8),
        ("count",    16),
        ("icc",       8),
        ("control",   8)
    ]
    payload_layout = [("data", dw)]
    return EndpointDescription(payload_layout, param_layout)


def transport_rx_description(dw):
    param_layout = [
        ("type",            8),
        ("pm_port",         4),
        ("r",               1),
        ("d",               1),
        ("i",               1),
        ("status",          8),
        ("errors",          8),
        ("lba",            48),
        ("device",          8),
        ("count",          16),
        ("transfer_count", 16),
        ("error",           1)
    ]
    payload_layout = [("data", dw)]
    return EndpointDescription(payload_layout, param_layout)


# Command Layer ------------------------------------------------------------------------------------

regs = {
    "WRITE_DMA_EXT":   0x35,
    "READ_DMA_EXT":    0x25,
    "IDENTIFY_DEVICE": 0xec
}

reg_d2h_status = {
    "bsy":  7,
    "drdy": 6,
    "df":   5,
    "se":   5,
    "dwe":  4,
    "drq":  3,
    "ae":   2,
    "sns":  1,
    "cc":   0,
    "err":  0
}

def command_tx_description(dw):
    param_layout = [
        ("write",    1),
        ("read",     1),
        ("identify", 1),
        ("sector",  48),
        ("count",   16)
    ]
    payload_layout = [("data", dw)]
    return EndpointDescription(payload_layout, param_layout)


def command_rx_description(dw):
    param_layout = [
        ("write",    1),
        ("read",     1),
        ("identify", 1),
        ("end",      1),
        ("failed",   1)
    ]
    payload_layout = [("data", dw)]
    return EndpointDescription(payload_layout, param_layout)


def command_rx_cmd_description(dw):
    param_layout = [
        ("write",    1),
        ("read",     1),
        ("identify", 1),
        ("end",      1),
        ("failed",   1)
    ]
    payload_layout = [("dummy", 1)]
    return EndpointDescription(payload_layout, param_layout)


def command_rx_data_description(dw):
    payload_layout = [("data", dw)]
    return EndpointDescription(payload_layout)


# HDD ----------------------------------------------------------------------------------------------

logical_sector_size = 512  # constant since all HDDs use this

def dwords2sectors(n):
    return ceil(n*4/logical_sector_size)


def sectors2dwords(n):
    return n*logical_sector_size//4



#
# This file is part of LiteSATA.
#
# Copyright (c) 2015-2020 Florent Kermarrec <florent@enjoy-digital.fr>
# Copyright (c) 2017 Johan Klockars <Johan.Klockars@hasselblad.com>
# Copyright (c) 2015-2016 Olof Kindgren <olof.kindgren@gmail.com>
# SPDX-License-Identifier: BSD-2-Clause

from litesata.common import *
from litesata.common import _PulseSynchronizer, _RisingEdge

from migen.genlib.cdc import MultiReg
from migen.genlib.resetsync import AsyncResetSynchronizer

from litex.soc.cores.clock import S7MMCM

from liteiclink.serdes.gtx_7series_init import GTXTXInit, GTXRXInit

# --------------------------------------------------------------------------------------------------

class K7LiteSATAPHYCRG(Module):
    def __init__(self, refclk, pads, gtx, gen, tx_buffer_enable=False):
        self.tx_reset = Signal()
        self.rx_reset = Signal()

        self.clock_domains.cd_sata_tx = ClockDomain()
        self.clock_domains.cd_sata_rx = ClockDomain()

        # CPLL -------------------------------------------------------------------------------------
        #   (gen3) 150MHz / VCO @ 3GHz / Linerate @ 6Gbps
        #   (gen2 & gen1) VCO still @ 3 GHz, Linerate is decreased with output dividers.
        if isinstance(refclk, (Signal, ClockSignal)):
            self.refclk = refclk
        else:
            self.refclk = Signal()
            self.specials += Instance("IBUFDS_GTE2",
                i_CEB = 0,
                i_I   = pads.clk_p,
                i_IB  = pads.clk_n,
                o_O   = self.refclk
            )
        self.comb += gtx.gtrefclk0.eq(self.refclk)

        # TX clocking ------------------------------------------------------------------------------
        #   (gen3) 150MHz from CPLL TXOUTCLK, sata_tx clk @ 300MHz (16-bits) /  150MHz (32-bits)
        #   (gen2) 150MHz from CPLL TXOUTCLK, sata_tx clk @ 150MHz (16-bits) /   75MHz (32-bits)
        #   (gen1) 150MHz from CPLL TXOUTCLK, sata_tx clk @ 75MHz  (16-bits) / 37.5MHz (32-bits)
        if tx_buffer_enable:
            self.specials += Instance("BUFG", i_I=gtp.txoutclk, o_O=self.cd_sata_tx.clk)
        else:
            tx_mmcm_clkin = Signal()
            tx_mmcm = S7MMCM(speedgrade=-1)
            tx_mmcm_clkout_freq = {
                "gen1":  75e6/(gtx.data_width/16),
                "gen2": 150e6/(gtx.data_width/16),
                "gen3": 300e6/(gtx.data_width/16),
            }
            self.submodules += tx_mmcm
            self.specials += Instance("BUFG", i_I=gtx.txoutclk, o_O=tx_mmcm_clkin)
            tx_mmcm.register_clkin(tx_mmcm_clkin, 150e6)
            tx_mmcm.create_clkout(self.cd_sata_tx, tx_mmcm_clkout_freq[gen], with_reset=False)

        self.comb += gtx.txusrclk.eq(self.cd_sata_tx.clk)
        self.comb += gtx.txusrclk2.eq(self.cd_sata_tx.clk)

        # RX clocking ------------------------------------------------------------------------------
        #   (gen3) sata_rx recovered clk @ 300MHz (16-bits) /  150MHz (32-bits) from GTX RXOUTCLK
        #   (gen2) sata_rx recovered clk @ 150MHz (16-bits) /   75MHz (32-bits) from GTX RXOUTCLK
        #   (gen1) sata_rx recovered clk @ 75MHz  (16-bits) / 37.5MHz (32-bits) from GTX RXOUTCLK
        self.specials += Instance("BUFG", i_I=gtx.rxoutclk, o_O=self.cd_sata_rx.clk)

        self.comb += gtx.rxusrclk.eq(self.cd_sata_rx.clk)
        self.comb += gtx.rxusrclk2.eq(self.cd_sata_rx.clk)

        # Reset for SATA TX/RX clock domains -------------------------------------------------------
        self.specials += [
            AsyncResetSynchronizer(self.cd_sata_tx, ~gtx.cplllock | self.tx_reset),
            AsyncResetSynchronizer(self.cd_sata_rx, ~gtx.cplllock | self.rx_reset)
        ]

# --------------------------------------------------------------------------------------------------

class K7LiteSATAPHY(Module):
    def __init__(self, pads, gen, clk_freq, data_width=16, tx_buffer_enable=False, rx_buffer_enable=False):
        assert data_width in [16, 32]
        # Common signals
        self.data_width     = data_width

        # Control
        self.ready          = Signal() # o

        self.tx_idle        = Signal() # i
        self.tx_polarity    = Signal() # i

        self.tx_cominit_stb = Signal() # i
        self.tx_cominit_ack = Signal() # o
        self.tx_comwake_stb = Signal() # i
        self.tx_comwake_ack = Signal() # o

        self.rx_idle        = Signal() # o
        self.rx_cdrhold     = Signal() # i
        self.rx_polarity    = Signal() # i

        self.rx_cominit_stb = Signal() # o
        self.rx_comwake_stb = Signal() # o

        self.rxdisperr      = Signal(data_width//8) # o
        self.rxnotintable   = Signal(data_width//8) # o

        # Datapath
        self.sink           = stream.Endpoint(phy_description(data_width))
        self.source         = stream.Endpoint(phy_description(data_width))

        # K7 specific signals
        # Channel - Ref Clock Ports
        self.gtrefclk0      = Signal()

        # Channel PLL
        self.cplllock       = Signal()

        # Receive Ports - 8b10b Decoder
        self.rxcharisk      = Signal(data_width//8)

        # Receive Ports - RX Data Path interface
        self.rxdata         = Signal(data_width)
        self.rxoutclk       = Signal()
        self.rxusrclk       = Signal()
        self.rxusrclk2      = Signal()

        # Receive Ports - RX Ports for SATA
        self.rxcominitdet   = Signal()
        self.rxcomwakedet   = Signal()

        # Transmit Ports - 8b10b Encoder Control Ports
        self.txcharisk      = Signal(data_width//8)

        # Transmit Ports - TX Data Path interface
        self.txdata         = Signal(data_width)
        self.txoutclk       = Signal()
        self.txusrclk       = Signal()
        self.txusrclk2      = Signal()

        # Transmit Ports - TX Ports for PCI Express
        self.txelecidle     = Signal(reset=1)

        # Transmit Ports - TX Ports for SATA
        self.txcomfinish    = Signal()
        self.txcominit      = Signal()
        self.txcomwake      = Signal()

        # Power-down signals
        self.cpllpd         = Signal()
        self.rxpd           = Signal()
        self.txpd           = Signal()

        # Config at startup
        div_config = {
            "gen1": 4,
            "gen2": 2,
            "gen3": 1,
        }
        rxout_div = div_config[gen]
        txout_div = div_config[gen]

        cdr_config = {
            "gen1": 0x0380008bff40100008,
            "gen2": 0x0388008bff40200008,
            "gen3": 0x0380008bff10200010,
        }
        rxcdr_cfg = cdr_config[gen]

        # TX Init ----------------------------------------------------------------------------------
        self.submodules.tx_init = tx_init = GTXTXInit(clk_freq, buffer_enable=tx_buffer_enable)
        self.comb += tx_init.plllock.eq(self.cplllock)

        # RX Init ----------------------------------------------------------------------------------
        self.submodules.rx_init = rx_init = GTXRXInit(clk_freq, buffer_enable=rx_buffer_enable)
        self.comb += rx_init.plllock.eq(self.cplllock)

        # Ready ------------------------------------------------------------------------------------
        self.comb += self.ready.eq(tx_init.done & rx_init.done)

        # Specific / Generic signals encoding/decoding ---------------------------------------------
        self.comb += [
            self.txelecidle.eq(self.tx_idle | self.txpd),
            self.tx_cominit_ack.eq(self.tx_cominit_stb & self.txcomfinish),
            self.tx_comwake_ack.eq(self.tx_comwake_stb & self.txcomfinish),
            self.rx_cominit_stb.eq(self.rxcominitdet),
            self.rx_comwake_stb.eq(self.rxcomwakedet),
        ]
        self.submodules += _RisingEdge(self.tx_cominit_stb, self.txcominit)
        self.submodules += _RisingEdge(self.tx_comwake_stb, self.txcomwake)

        self.sync.sata_rx += [
            self.source.valid.eq(1),
            self.source.charisk.eq(self.rxcharisk),
            self.source.data.eq(self.rxdata)
        ]

        self.sync.sata_tx += [
            self.txcharisk.eq(self.sink.charisk),
            self.txdata.eq(self.sink.data),
            self.sink.ready.eq(1),
        ]

        # Internals and clock domain crossing ------------------------------------------------------
        # sys_clk --> sata_tx clk
        txpd       = Signal()
        txelecidle = Signal(reset=1)
        txcominit  = Signal()
        txcomwake  = Signal()
        self.specials += [
            MultiReg(self.txpd,             txpd, "sata_tx"),
            MultiReg(self.txelecidle, txelecidle, "sata_tx"),
        ]
        self.submodules += [
            _PulseSynchronizer(self.txcominit,  "sys",  txcominit, "sata_tx"),
            _PulseSynchronizer(self.txcomwake,  "sys",  txcomwake, "sata_tx"),
        ]

        # sata_tx clk --> sys clk
        txcomfinish = Signal()
        self.submodules += _PulseSynchronizer(txcomfinish, "sata_tx", self.txcomfinish, "sys")

        # sata_rx clk --> sys clk
        rxcominitdet = Signal()
        rxcomwakedet = Signal()
        rxratedone   = Signal()
        rxdisperr    = Signal(data_width//8)
        rxnotintable = Signal(data_width//8)
        self.specials += [
            MultiReg(rxcominitdet, self.rxcominitdet, "sys"),
            MultiReg(rxcomwakedet, self.rxcomwakedet, "sys"),
            MultiReg(rxdisperr, self.rxdisperr, "sys"),
            MultiReg(rxnotintable, self.rxnotintable, "sys")
        ]

        # QPLL input clock -------------------------------------------------------------------------
        self.qpllclk    = Signal()
        self.qpllrefclk = Signal()

        # OOB clock (75MHz) ------------------------------------------------------------------------
        oobclk = Signal()
        self.specials += Instance("FDPE",
            p_INIT = 1,
            i_CE   = 1,
            i_PRE  = 0,
            i_C    = self.gtrefclk0,
            i_D    = ~oobclk,
            o_Q    = oobclk
        )

        # GTXE2_CHANNEL Instance -------------------------------------------------------------------
        class Open(Signal): pass
        gtx_params = dict(
            # Simulation-Only Attributes
            p_SIM_RECEIVER_DETECT_PASS     = "TRUE",
            p_SIM_TX_EIDLE_DRIVE_LEVEL     = "X",
            p_SIM_RESET_SPEEDUP            = "FALSE",
            p_SIM_CPLLREFCLK_SEL           = "FALSE",
            p_SIM_VERSION                  = "4.0",

            # RX Byte and Word Alignment Attributes
            p_ALIGN_COMMA_DOUBLE           = "FALSE",
            p_ALIGN_COMMA_ENABLE           = 0b1111111111,
            p_ALIGN_COMMA_WORD             = 2 if data_width == 16 else 4,
            p_ALIGN_MCOMMA_DET             = "TRUE",
            p_ALIGN_MCOMMA_VALUE           = 0b1010000011,
            p_ALIGN_PCOMMA_DET             = "TRUE",
            p_ALIGN_PCOMMA_VALUE           = 0b0101111100,
            p_SHOW_REALIGN_COMMA           = "TRUE",
            p_RXSLIDE_AUTO_WAIT            = 7,
            p_RXSLIDE_MODE                 = "OFF" if rx_buffer_enable else "PCS",
            p_RX_SIG_VALID_DLY             = 10,

            # RX 8B/10B Decoder Attributes
            p_RX_DISPERR_SEQ_MATCH         = "TRUE",
            p_DEC_MCOMMA_DETECT            = "TRUE",
            p_DEC_PCOMMA_DETECT            = "TRUE",
            p_DEC_VALID_COMMA_ONLY         = "TRUE",

            # RX Clock Correction Attributes
            p_CBCC_DATA_SOURCE_SEL         = "DECODED",
            p_CLK_COR_SEQ_2_USE            = "FALSE",
            p_CLK_COR_KEEP_IDLE            = "FALSE",
            p_CLK_COR_MAX_LAT              = 9 if data_width == 16 else 20,
            p_CLK_COR_MIN_LAT              = 7 if data_width == 16 else 16,
            p_CLK_COR_PRECEDENCE           = "TRUE",
            p_CLK_COR_REPEAT_WAIT          = 0,
            p_CLK_COR_SEQ_LEN              = 1,
            p_CLK_COR_SEQ_1_ENABLE         = 0b1111,
            p_CLK_COR_SEQ_1_1              = 0b0100000000,
            p_CLK_COR_SEQ_1_2              = 0b0000000000,
            p_CLK_COR_SEQ_1_3              = 0b0000000000,
            p_CLK_COR_SEQ_1_4              = 0b0000000000,
            p_CLK_CORRECT_USE              = "FALSE",
            p_CLK_COR_SEQ_2_ENABLE         = 0b1111,
            p_CLK_COR_SEQ_2_1              = 0b0100000000,
            p_CLK_COR_SEQ_2_2              = 0b0000000000,
            p_CLK_COR_SEQ_2_3              = 0b0000000000,
            p_CLK_COR_SEQ_2_4              = 0b0000000000,

            # RX Channel Bonding Attributes
            p_CHAN_BOND_KEEP_ALIGN         = "FALSE",
            p_CHAN_BOND_MAX_SKEW           = 1,
            p_CHAN_BOND_SEQ_LEN            = 1,
            p_CHAN_BOND_SEQ_1_1            = 0b0000000000,
            p_CHAN_BOND_SEQ_1_2            = 0b0000000000,
            p_CHAN_BOND_SEQ_1_3            = 0b0000000000,
            p_CHAN_BOND_SEQ_1_4            = 0b0000000000,
            p_CHAN_BOND_SEQ_1_ENABLE       = 0b1111,
            p_CHAN_BOND_SEQ_2_1            = 0b0000000000,
            p_CHAN_BOND_SEQ_2_2            = 0b0000000000,
            p_CHAN_BOND_SEQ_2_3            = 0b0000000000,
            p_CHAN_BOND_SEQ_2_4            = 0b0000000000,
            p_CHAN_BOND_SEQ_2_ENABLE       = 0b1111,
            p_CHAN_BOND_SEQ_2_USE          = "FALSE",
            p_FTS_DESKEW_SEQ_ENABLE        = 0b1111,
            p_FTS_LANE_DESKEW_CFG          = 0b1111,
            p_FTS_LANE_DESKEW_EN           = "FALSE",

            # RX Margin Analysis Attributes
            p_ES_CONTROL                   = 0b000000,
            p_ES_ERRDET_EN                 = "FALSE",
            p_ES_EYE_SCAN_EN               = "TRUE",
            p_ES_HORZ_OFFSET               = 0x000,
            p_ES_PMA_CFG                   = 0b0000000000,
            p_ES_PRESCALE                  = 0b00000,
            p_ES_QUALIFIER                 = 0x00000000000000000000,
            p_ES_QUAL_MASK                 = 0x00000000000000000000,
            p_ES_SDATA_MASK                = 0x00000000000000000000,
            p_ES_VERT_OFFSET               = 0b000000000,

            # FPGA RX Interface Attributes
            p_RX_DATA_WIDTH                = 20 if data_width == 16 else 40,

            # PMA Attributes
            p_OUTREFCLK_SEL_INV            = 0b11,
            p_PMA_RSV                      = 0x00018480,
            p_PMA_RSV2                     = 0x2050,
            p_PMA_RSV3                     = 0b00,
            p_PMA_RSV4                     = 0x00000000,
            p_RX_BIAS_CFG                  = 0b000000000100,
            p_DMONITOR_CFG                 = 0x000A00,
            p_RX_CM_SEL                    = 0b11,
            p_RX_CM_TRIM                   = 0b010,
            p_RX_DEBUG_CFG                 = 0b000000000000,
            p_RX_OS_CFG                    = 0b0000010000000,
            p_TERM_RCAL_CFG                = 0b10000,
            p_TERM_RCAL_OVRD               = 0b0,
            p_TST_RSV                      = 0x00000000,
            p_RX_CLK25_DIV                 = 6,
            p_TX_CLK25_DIV                 = 6,
            p_UCODEER_CLR                  = 0b0,

            # PCI Express Attributes
            p_PCS_PCIE_EN                  = "FALSE",

            # PCS Attributes
            p_PCS_RSVD_ATTR                = 0x108 if gen == "gen1" else 0x100,

            # RX Buffer Attributes
            p_RXBUF_ADDR_MODE              = "FAST",
            p_RXBUF_EIDLE_HI_CNT           = 0b1000,
            p_RXBUF_EIDLE_LO_CNT           = 0b0000,
            p_RXBUF_EN                     = "TRUE" if rx_buffer_enable else "FALSE",
            p_RX_BUFFER_CFG                = 0b000000,
            p_RXBUF_RESET_ON_CB_CHANGE     = "TRUE",
            p_RXBUF_RESET_ON_COMMAALIGN    = "FALSE",
            p_RXBUF_RESET_ON_EIDLE         = "FALSE",
            p_RXBUF_RESET_ON_RATE_CHANGE   = "TRUE",
            p_RXBUFRESET_TIME              = 0b00001,
            p_RXBUF_THRESH_OVFLW           = 61,
            p_RXBUF_THRESH_OVRD            = "FALSE",
            p_RXBUF_THRESH_UNDFLW          = 4,
            p_RXDLY_CFG                    = 0x001F,
            p_RXDLY_LCFG                   = 0x030,
            p_RXDLY_TAP_CFG                = 0x0000,
            p_RXPH_CFG                     = 0x000000,
            p_RXPHDLY_CFG                  = 0x084020,
            p_RXPH_MONITOR_SEL             = 0b00000,
            p_RX_XCLK_SEL                  = "RXREC" if rx_buffer_enable else "RXUSR",
            p_RX_DDI_SEL                   = 0b000000,
            p_RX_DEFER_RESET_BUF_EN        = "TRUE",

            # CDR Attributes
            p_RXCDR_CFG                    = rxcdr_cfg,
            p_RXCDR_FR_RESET_ON_EIDLE      = 0b0,
            p_RXCDR_HOLD_DURING_EIDLE      = 0b0,
            p_RXCDR_PH_RESET_ON_EIDLE      = 0b0,
            p_RXCDR_LOCK_CFG               = 0b010101,

            # RX Initialization and Reset Attributes
            p_RXCDRFREQRESET_TIME          = 0b00001,
            p_RXCDRPHRESET_TIME            = 0b00001,
            p_RXISCANRESET_TIME            = 0b00001,
            p_RXPCSRESET_TIME              = 0b00001,
            p_RXPMARESET_TIME              = 0b00011,

            # RX OOB Signaling Attributes
            p_RXOOB_CFG                    = 0b0000110,

            # RX Gearbox Attributes
            p_RXGEARBOX_EN                 = "FALSE",
            p_GEARBOX_MODE                 = 0b000,

            # PRBS Detection Attribute
            p_RXPRBS_ERR_LOOPBACK          = 0b0,

            # Power-Down Attributes
            p_PD_TRANS_TIME_FROM_P2        = 0x03c,
            p_PD_TRANS_TIME_NONE_P2        = 0x3c,
            p_PD_TRANS_TIME_TO_P2          = 0x64,

            # RX OOB Signaling Attributes
            p_SAS_MAX_COM                  = 64,
            p_SAS_MIN_COM                  = 36,
            p_SATA_BURST_SEQ_LEN           = 0b0101,
            p_SATA_BURST_VAL               = 0b100,
            p_SATA_EIDLE_VAL               = 0b100,
            p_SATA_MAX_BURST               = 8,
            p_SATA_MAX_INIT                = 21,
            p_SATA_MAX_WAKE                = 7,
            p_SATA_MIN_BURST               = 4,
            p_SATA_MIN_INIT                = 12,
            p_SATA_MIN_WAKE                = 4,

            # RX Fabric Clock Output Control Attributes
            p_TRANS_TIME_RATE              = 0x0E,

            # TX Buffer Attributes
            p_TXBUF_EN                     = "TRUE" if tx_buffer_enable else "FALSE",
            p_TXBUF_RESET_ON_RATE_CHANGE   = "TRUE",
            p_TXDLY_CFG                    = 0x001F,
            p_TXDLY_LCFG                   = 0x030,
            p_TXDLY_TAP_CFG                = 0x0000,
            p_TXPH_CFG                     = 0x0780,
            p_TXPHDLY_CFG                  = 0x084020,
            p_TXPH_MONITOR_SEL             = 0b00000,
            p_TX_XCLK_SEL                  = "TXOUT" if tx_buffer_enable else "TXUSR",

            # FPGA TX Interface Attributes
            p_TX_DATA_WIDTH                = 20 if data_width == 16 else 40,

            # TX Configurable Driver Attributes
            p_TX_DEEMPH0                   = 0b00000,
            p_TX_DEEMPH1                   = 0b00000,
            p_TX_EIDLE_ASSERT_DELAY        = 0b110,
            p_TX_EIDLE_DEASSERT_DELAY      = 0b100,
            p_TX_LOOPBACK_DRIVE_HIZ        = "FALSE",
            p_TX_MAINCURSOR_SEL            = 0b0,
            p_TX_DRIVE_MODE                = "DIRECT",
            p_TX_MARGIN_FULL_0             = 0b1001110,
            p_TX_MARGIN_FULL_1             = 0b1001001,
            p_TX_MARGIN_FULL_2             = 0b1000101,
            p_TX_MARGIN_FULL_3             = 0b1000010,
            p_TX_MARGIN_FULL_4             = 0b1000000,
            p_TX_MARGIN_LOW_0              = 0b1000110,
            p_TX_MARGIN_LOW_1              = 0b1000100,
            p_TX_MARGIN_LOW_2              = 0b1000010,
            p_TX_MARGIN_LOW_3              = 0b1000000,
            p_TX_MARGIN_LOW_4              = 0b1000000,

            # TX Gearbox Attributes
            p_TXGEARBOX_EN                 = "FALSE",

            # TX Initialization and Reset Attributes
            p_TXPCSRESET_TIME              = 0b00001,
            p_TXPMARESET_TIME              = 0b00001,

            # TX Receiver Detection Attributes
            p_TX_RXDETECT_CFG              = 0x1832,
            p_TX_RXDETECT_REF              = 0b100,

            # CPLL Attributes
            p_CPLL_CFG                     = 0xBC07DC,
            p_CPLL_FBDIV                   = 4,
            p_CPLL_FBDIV_45                = 5,
            p_CPLL_INIT_CFG                = 0x00001E,
            p_CPLL_LOCK_CFG                = 0x01E8,
            p_CPLL_REFCLK_DIV              = 1,
            p_RXOUT_DIV                    = rxout_div,
            p_TXOUT_DIV                    = txout_div,
            p_SATA_CPLL_CFG                = "VCO_3000MHZ",

            # RX Initialization and Reset Attributes
            p_RXDFELPMRESET_TIME           = 0b0001111,

            # RX Equalizer Attributes
            p_RXLPM_HF_CFG                 = 0b00000011110000,
            p_RXLPM_LF_CFG                 = 0b00000011110000,
            p_RX_DFE_GAIN_CFG              = 0x020FEA,
            p_RX_DFE_H2_CFG                = 0b000000000000,
            p_RX_DFE_H3_CFG                = 0b000001000000,
            p_RX_DFE_H4_CFG                = 0b00011110000,
            p_RX_DFE_H5_CFG                = 0b00011100000,
            p_RX_DFE_KL_CFG                = 0b0000011111110,
            p_RX_DFE_LPM_CFG               = 0x0954,
            p_RX_DFE_LPM_HOLD_DURING_EIDLE = 0b1,
            p_RX_DFE_UT_CFG                = 0b10001111000000000,
            p_RX_DFE_VP_CFG                = 0b00011111100000011,

            # Power-Down Attributes
            p_RX_CLKMUX_PD                 = 0b1,
            p_TX_CLKMUX_PD                 = 0b1,

            # FPGA RX Interface Attribute
            p_RX_INT_DATAWIDTH             = data_width == 32,

            # FPGA TX Interface Attribute
            p_TX_INT_DATAWIDTH             = data_width == 32,

            # TX Configurable Driver Attributes
            p_TX_QPI_STATUS_EN             = 0b0,

            # RX Equalizer Attributes
            p_RX_DFE_KL_CFG2               = 0x3310180c,
            p_RX_DFE_XYD_CFG               = 0b0000000000000,

            # TX Configurable Driver Attributes
            p_TX_PREDRIVER_MODE            = 0b0
        )
        gtx_params.update(
            # CPLL Ports
            o_CPLLFBCLKLOST    = Open(),
            o_CPLLLOCK         = self.cplllock,
            i_CPLLLOCKDETCLK   = 0,
            i_CPLLLOCKEN       = 1,
            i_CPLLPD           = self.cpllpd,
            o_CPLLREFCLKLOST   = Open(),
            i_CPLLREFCLKSEL    = 0b001,
            i_CPLLRESET        = tx_init.pllreset,
            i_GTRSVD           = 0b0000000000000000,
            i_PCSRSVDIN        = 0b0000000000000000,
            i_PCSRSVDIN2       = 0b00000,
            i_PMARSVDIN        = 0b00000,
            i_PMARSVDIN2       = 0b00000,
            i_TSTIN            = 0b11111111111111111111,
            o_TSTOUT           = Open(),

            # Channel
            i_CLKRSVD          = oobclk,

            # Channel - Clocking Ports
            i_GTGREFCLK        = 0,
            i_GTNORTHREFCLK0   = 0,
            i_GTNORTHREFCLK1   = 0,
            i_GTREFCLK0        = self.gtrefclk0,
            i_GTREFCLK1        = 0,
            i_GTSOUTHREFCLK0   = 0,
            i_GTSOUTHREFCLK1   = 0,

            # Channel - DRP Ports
            i_DRPADDR          = 0,
            i_DRPCLK           = 0,
            i_DRPDI            = 0,
            o_DRPDO            = Open(),
            i_DRPEN            = 0,
            o_DRPRDY           = Open(),
            i_DRPWE            = 0,

            # Clocking Ports
            o_GTREFCLKMONITOR  = Open(),
            i_QPLLCLK          = self.qpllclk,
            i_QPLLREFCLK       = self.qpllrefclk,
            i_RXSYSCLKSEL      = 0b00,
            i_TXSYSCLKSEL      = 0b00,

            # Digital Monitor Ports
            o_DMONITOROUT      = Open(),

            # FPGA TX Interface Datapath Configuration
            i_TX8B10BEN        = 1,

            # Loopback Ports
            i_LOOPBACK         = 0,

            # PCI Express Ports
            o_PHYSTATUS        = Open(),
            i_RXRATE           = 0b000,
            o_RXVALID          = Open(),

            # Power-Down Ports
            i_RXPD             = Replicate(self.rxpd, 2),
            i_TXPD             = Replicate(txpd, 2),

            # RX 8B/10B Decoder Ports
            i_SETERRSTATUS     = 0,

            # RX Initialization and Reset Ports
            i_EYESCANRESET     = 0,
            i_RXUSERRDY        = rx_init.Xxuserrdy,

            # RX Margin Analysis Ports
            o_EYESCANDATAERROR = Open(),
            i_EYESCANMODE      = 0,
            i_EYESCANTRIGGER   = 0,

            # Receive Ports - CDR Ports
            i_RXCDRFREQRESET   = 0,
            i_RXCDRHOLD        = self.rx_cdrhold,
            o_RXCDRLOCK        = Open(),
            i_RXCDROVRDEN      = 0,
            i_RXCDRRESET       = 0,
            i_RXCDRRESETRSV    = 0,

            # Receive Ports - Clock Correction Ports
            o_RXCLKCORCNT      = Open(),

            # Receive Ports - FPGA RX Interface Datapath Configuration
            i_RX8B10BEN        = 1,

            # Receive Ports - FPGA RX Interface Ports
            i_RXUSRCLK         = self.rxusrclk,
            i_RXUSRCLK2        = self.rxusrclk2,

            # Receive Ports - FPGA RX interface Ports
            o_RXDATA           = self.rxdata,

            # Receive Ports - Pattern Checker Ports
            o_RXPRBSERR        = Open(),
            i_RXPRBSSEL        = 0b000,

            # Receive Ports - Pattern Checker ports
            i_RXPRBSCNTRESET   = 0,

            # Receive Ports - RX  Equalizer Ports
            i_RXDFEXYDEN       = 1,
            i_RXDFEXYDHOLD     = 0,
            i_RXDFEXYDOVRDEN   = 0,

            # Receive Ports - RX 8B/10B Decoder Ports
            o_RXDISPERR        = rxdisperr,
            o_RXNOTINTABLE     = rxnotintable,

            # Receive Ports - RX AFE
            i_GTXRXP           = pads.rx_p,
            i_GTXRXN           = pads.rx_n,

            # Receive Ports - RX Buffer Bypass Ports
            i_RXBUFRESET       = 0,
            o_RXBUFSTATUS      = Open(),
            i_RXDDIEN          = 0 if rx_buffer_enable else 1,
            i_RXDLYBYPASS      = 1 if rx_buffer_enable else 0,
            i_RXDLYEN          = 0,
            i_RXDLYOVRDEN      = 0,
            i_RXDLYSRESET      = rx_init.Xxdlysreset,
            o_RXDLYSRESETDONE  = rx_init.Xxdlysresetdone,
            i_RXPHALIGN        = 0,
            o_RXPHALIGNDONE    = rx_init.Xxphaligndone,
            i_RXPHALIGNEN      = 0,
            i_RXPHDLYPD        = 0,
            i_RXPHDLYRESET     = 0,
            o_RXPHMONITOR      = Open(),
            i_RXPHOVRDEN       = 0,
            o_RXPHSLIPMONITOR  = Open(),
            o_RXSTATUS         = Open(),

            # Receive Ports - RX Byte and Word Alignment Ports
            o_RXBYTEISALIGNED  = Open(),
            o_RXBYTEREALIGN    = Open(),
            o_RXCOMMADET       = Open(),
            i_RXCOMMADETEN     = 1,
            i_RXMCOMMAALIGNEN  = 1,
            i_RXPCOMMAALIGNEN  = 1,

            # Receive Ports - RX Channel Bonding Ports
            o_RXCHANBONDSEQ    = Open(),
            i_RXCHBONDEN       = 0,
            i_RXCHBONDLEVEL    = 0b000,
            i_RXCHBONDMASTER   = 0,
            o_RXCHBONDO        = Open(),
            i_RXCHBONDSLAVE    = 0,

            # Receive Ports - RX Channel Bonding Ports
            o_RXCHANISALIGNED  = Open(),
            o_RXCHANREALIGN    = Open(),

            # Receive Ports - RX Equailizer Ports
            i_RXLPMHFHOLD      = 0,
            i_RXLPMHFOVRDEN    = 0,
            i_RXLPMLFHOLD      = 0,

            # Receive Ports - RX Equalizer Ports
            i_RXDFEAGCHOLD     = 0,
            i_RXDFEAGCOVRDEN   = 0,
            i_RXDFECM1EN       = 0,
            i_RXDFELFHOLD      = 0,
            i_RXDFELFOVRDEN    = 0,
            i_RXDFELPMRESET    = 0,
            i_RXDFETAP2HOLD    = 0,
            i_RXDFETAP2OVRDEN  = 0,
            i_RXDFETAP3HOLD    = 0,
            i_RXDFETAP3OVRDEN  = 0,
            i_RXDFETAP4HOLD    = 0,
            i_RXDFETAP4OVRDEN  = 0,
            i_RXDFETAP5HOLD    = 0,
            i_RXDFETAP5OVRDEN  = 0,
            i_RXDFEUTHOLD      = 0,
            i_RXDFEUTOVRDEN    = 0,
            i_RXDFEVPHOLD      = 0,
            i_RXDFEVPOVRDEN    = 0,
            i_RXDFEVSEN        = 0,
            i_RXLPMLFKLOVRDEN  = 0,
            o_RXMONITOROUT     = Open(),
            i_RXMONITORSEL     = 0,
            i_RXOSHOLD         = 0,
            i_RXOSOVRDEN       = 0,

            # Receive Ports - RX Fabric ClocK Output Control Ports
            o_RXRATEDONE       = Open(),

            # Receive Ports - RX Fabric Output Control Ports
            o_RXOUTCLK         = self.rxoutclk,
            o_RXOUTCLKFABRIC   = Open(),
            o_RXOUTCLKPCS      = Open(),
            i_RXOUTCLKSEL      = 0b010,

            # Receive Ports - RX Gearbox Ports
            o_RXDATAVALID      = Open(),
            o_RXHEADER         = Open(),
            o_RXHEADERVALID    = Open(),
            o_RXSTARTOFSEQ     = Open(),

            # Receive Ports - RX Gearbox Ports
            i_RXGEARBOXSLIP    = 0,

            # Receive Ports - RX Initialization and Reset Ports
            i_GTRXRESET        = rx_init.gtXxreset,
            i_RXOOBRESET       = 0,
            i_RXPCSRESET       = 0,
            i_RXPMARESET       = 0,

            # Receive Ports - RX Margin Analysis ports
            i_RXLPMEN          = 1,

            # Receive Ports - RX OOB Signaling ports
            o_RXCOMSASDET      = Open(),
            o_RXCOMWAKEDET     = rxcomwakedet,

            # Receive Ports - RX OOB Signaling ports
            o_RXCOMINITDET     = rxcominitdet,

            # Receive Ports - RX OOB signalling Ports
            o_RXELECIDLE       = Open(),
            i_RXELECIDLEMODE   = 0b00,

            # Receive Ports - RX Polarity Control Ports
            i_RXPOLARITY       = self.rx_polarity,

            # Receive Ports - RX gearbox ports
            i_RXSLIDE          = 0,

            # Receive Ports - RX8B/10B Decoder Ports
            o_RXCHARISCOMMA    = Open(),
            o_RXCHARISK        = self.rxcharisk,

            # Receive Ports - Rx Channel Bonding Ports
            i_RXCHBONDI        = 0b00000,

            # Receive Ports -RX Initialization and Reset Ports
            o_RXRESETDONE      = rx_init.Xxresetdone,

            # Rx AFE Ports
            i_RXQPIEN          = 0,
            o_RXQPISENN        = Open(),
            o_RXQPISENP        = Open(),

            # TX Buffer Bypass Ports
            i_TXPHDLYTSTCLK    = 0,

            # TX Configurable Driver Ports
            i_TXPOSTCURSOR     = 0b00000,
            i_TXPOSTCURSORINV  = 0,
            i_TXPRECURSOR      = 0b00000,
            i_TXPRECURSORINV   = 0,
            i_TXQPIBIASEN      = 0,
            i_TXQPISTRONGPDOWN = 0,
            i_TXQPIWEAKPUP     = 0,

            # TX Initialization and Reset Ports
            i_CFGRESET         = 0,
            i_GTTXRESET        = tx_init.gtXxreset,
            o_PCSRSVDOUT       = Open(),
            i_TXUSERRDY        = tx_init.Xxuserrdy,

            # Transceiver Reset Mode Operation
            i_GTRESETSEL       = 0,
            i_RESETOVRD        = 0,

            # Transmit Ports - 8b10b Encoder Control Ports
            i_TXCHARDISPMODE   = 0,
            i_TXCHARDISPVAL    = 0,

            # Transmit Ports - FPGA TX Interface Ports
            i_TXUSRCLK         = self.txusrclk,
            i_TXUSRCLK2        = self.txusrclk2,

            # Transmit Ports - PCI Express Ports
            i_TXELECIDLE       = txelecidle,
            i_TXMARGIN         = 0b000,
            i_TXRATE           = 0b000,
            i_TXSWING          = 0,

            # Transmit Ports - Pattern Generator Ports
            i_TXPRBSFORCEERR   = 0,

            # Transmit Ports - TX Buffer Bypass Ports
            i_TXDLYBYPASS      = 0,
            i_TXDLYEN          = 0,
            i_TXDLYHOLD        = 0,
            i_TXDLYOVRDEN      = 0,
            i_TXDLYSRESET      = tx_init.Xxdlysreset,
            o_TXDLYSRESETDONE  = tx_init.Xxdlysresetdone,
            i_TXDLYUPDOWN      = 0,
            i_TXPHALIGN        = 0,
            o_TXPHALIGNDONE    = tx_init.Xxphaligndone,
            i_TXPHALIGNEN      = 0,
            i_TXPHDLYPD        = 0,
            i_TXPHDLYRESET     = 0,
            i_TXPHINIT         = 0,
            o_TXPHINITDONE     = Open(),
            i_TXPHOVRDEN       = 0,

            # Transmit Ports - TX Buffer Ports
            o_TXBUFSTATUS      = Open(),

            # Transmit Ports - TX Configurable Driver Ports
            i_TXBUFDIFFCTRL    = 0b100,
            i_TXDEEMPH         = 0,
            i_TXDIFFCTRL       = 0b1000,
            i_TXDIFFPD         = 0,
            i_TXINHIBIT        = 0,
            i_TXMAINCURSOR     = 0b0000000,
            i_TXPISOPD         = 0,

            # Transmit Ports - TX Data Path interface
            i_TXDATA           = self.txdata,

            # Transmit Ports - TX Driver and OOB signaling
            o_GTXTXP           = pads.tx_p,
            o_GTXTXN           = pads.tx_n,

            # Transmit Ports - TX Fabric Clock Output Control Ports
            o_TXOUTCLK         = self.txoutclk,
            o_TXOUTCLKFABRIC   = Open(),
            o_TXOUTCLKPCS      = Open(),
            i_TXOUTCLKSEL      = 0b010 if tx_buffer_enable else 0b011 if gen == "gen2" else 0b100,
            o_TXRATEDONE       = Open(),

            # Transmit Ports - TX Gearbox Ports
            i_TXCHARISK        = self.txcharisk,
            o_TXGEARBOXREADY   = Open(),
            i_TXHEADER         = 0b000,
            i_TXSEQUENCE       = 0b0000000,
            i_TXSTARTSEQ       = 0,

            # Transmit Ports - TX Initialization and Reset Ports
            i_TXPCSRESET       = 0,
            i_TXPMARESET       = 0,
            o_TXRESETDONE      = tx_init.Xxresetdone,

            # Transmit Ports - TX OOB signaling Ports
            o_TXCOMFINISH      = txcomfinish,
            i_TXCOMINIT        = txcominit,
            i_TXCOMSAS         = 0,
            i_TXCOMWAKE        = txcomwake,
            i_TXPDELECIDLEMODE = 0,

            # Transmit Ports - TX Polarity Control Ports
            i_TXPOLARITY       = self.tx_polarity,

            # Transmit Ports - TX Receiver Detection Ports
            i_TXDETECTRX       = 0,

            # Transmit Ports - TX8b/10b Encoder Ports
            i_TX8B10BBYPASS    = 0b00000000,

            # Transmit Ports - pattern Generator Ports
            i_TXPRBSSEL        = 0b000,

            # Tx Configurable Driver  Ports
            o_TXQPISENN        = Open(),
            o_TXQPISENP        = Open(),
        )
        self.specials += Instance("GTXE2_CHANNEL", **gtx_params)



#
# This file is part of LiteSATA.
#
# Copyright (c) 2015-2019 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

from litesata.common import *

from migen.genlib.roundrobin import *

# LiteSATAMasterPort -------------------------------------------------------------------------------

class LiteSATAMasterPort:
    """SATA master port

    A port encapsulates 2 endpoints: A sink and a source.
    The sink if used to send command to the device an write data.
    The source is used to receive commands and data from the device.

    A master port drive command_tx and receive on command_rx.
    """
    def __init__(self, dw):
        self.dw     = dw
        self.source = stream.Endpoint(command_tx_description(dw))
        self.sink   = stream.Endpoint(command_rx_description(dw))

    def connect(self, slave):
        return [self.source.connect(slave.sink), slave.source.connect(self.sink)]

# LiteSATASlavePort --------------------------------------------------------------------------------

class LiteSATASlavePort:
    """SATA slave port

    A port encapsulates 2 endpoints: A sink and a source.
    The sink if used to send command to the device an write data.
    The source is used to receive commands and data from the device.

    A master port drive command_rx and receive on command_tx.
    """
    def __init__(self, dw):
        self.dw     = dw
        self.sink   = stream.Endpoint(command_tx_description(dw))
        self.source = stream.Endpoint(command_rx_description(dw))

    def connect(self, master):
        return [self.sink.connect(master.source), master.sink.connect(self.source)]

# LiteSATAUserPort ---------------------------------------------------------------------------------

class LiteSATAUserPort(LiteSATASlavePort):
    """SATA user port

    A user port is simply a slave port exposed to the user.
    """
    def __init__(self, dw, controller_dw=None):
        self.controller_dw = dw if controller_dw is None else controller_dw
        LiteSATASlavePort.__init__(self, dw)

# LiteSATAArbiter ----------------------------------------------------------------------------------

class LiteSATAArbiter(Module):
    """SATA arbiter

    Round robin arbiter between SATA user ports.
    """
    def __init__(self, users, master):
        self.rr = RoundRobin(len(users))
        self.submodules += self.rr
        self.grant = self.rr.grant
        cases = {}
        for i, slave in enumerate(users):
            sink, source = slave.sink, slave.source
            done    = Signal()
            ongoing = Signal()
            self.comb += done.eq(source.valid & source.last & source.end & source.ready)
            self.sync += \
                If(done,
                    ongoing.eq(0)
                ).Elif(sink.valid,
                    ongoing.eq(1)
                )
            self.comb += self.rr.request[i].eq((sink.valid | ongoing) & ~done)
            cases[i] = [users[i].connect(master)]
        self.comb += Case(self.grant, cases)

# LiteSATACrossbar ---------------------------------------------------------------------------------

class LiteSATACrossbar(Module):
    """SATA Crossbar

    This module provides a crossbar between a SATA master port from a controller
    and user ports requested with the get_port method.

    The controller be a PHY but also a RAID (Striping, Mirroring) module.
    """
    def __init__(self, controller):
        self.dw     = len(controller.sink.data)
        self.users  = []
        self.master = LiteSATAMasterPort(self.dw)
        self.comb += [
            self.master.source.connect(controller.sink),
            controller.source.connect(self.master.sink)
        ]

    def get_port(self, dw=None):
        dw = self.dw if dw is None else dw
        user_port     = LiteSATAUserPort(dw, self.dw)
        internal_port = LiteSATAUserPort(self.dw, self.dw)

        if dw != self.dw:
            converter = StrideConverter(
                command_tx_description(user_port.dw),
                command_tx_description(self.dw))
            self.submodules += converter
            self.comb += [
                user_port.sink.connect(converter.sink),
                converter.source.connect(internal_port.sink)
            ]

            converter = StrideConverter(
                command_rx_description(self.dw),
                command_rx_description(user_port.dw))
            self.submodules += converter
            self.comb += [
                internal_port.source.connect(converter.sink),
                converter.source.connect(user_port.source)
            ]

            self.users += [internal_port]
        else:
            self.users += [user_port]

        return user_port

    def get_ports(self, n, dw=None):
        dw    = self.dw if dw is None else dw
        ports = []
        for i in range(n):
            ports.append(self.get_port(dw))
        return ports

    def do_finalize(self):
        arbiter = LiteSATAArbiter(self.users, self.master)
        self.submodules += arbiter



#
# This file is part of LiteSATA.
#
# Copyright (c) 2015-2020 Florent Kermarrec <florent@enjoy-digital.fr>
# Copyright (c) 2017 Johan Klockars <Johan.Klockars@hasselblad.com>
# Copyright (c) 2016 Olof Kindgren <olof.kindgren@gmail.com>
# SPDX-License-Identifier: BSD-2-Clause

from litesata.common import *

from litex.gen.genlib.misc import WaitTimer

# LiteSATAPHYCtrl ----------------------------------------------------------------------------------

class LiteSATAPHYCtrl(Module):
    """SATA PHY Controller

    Manages link reset/initialization and OOB sequence.

    This modules is mainly a state machine that:
    - Commands and checks transceivers reset/initialization.
    - Manages the SATA OOB sequence with the device.
    - Wait for link to be stable before declaring it ready.
    - Re-initialize the link when invalid data are received.

    The state machine is robust enough to handle hot plug/ power off/on sequences of the device
    # without reseting the FPGA core.
    """
    def __init__(self, trx, crg, clk_freq):
        self.clk_freq = clk_freq
        self.ready    = Signal()
        self.sink     = sink   = stream.Endpoint(phy_description(32))
        self.source   = source = stream.Endpoint(phy_description(32))
        self.misalign = Signal()
        self.tx_idle  = Signal()
        self.rx_reset = Signal()
        self.tx_reset = Signal()
        self.rx_idle  = Signal()

        # # #

        # Always transmitting / receiving
        self.comb += source.valid.eq(1)
        self.comb += sink.ready.eq(1)

        # Retry / Align count/timers.
        retry_timer = WaitTimer(self.us(10000))
        align_timer = WaitTimer(self.us(873))
        align_count = Signal(4)
        self.submodules += align_timer, retry_timer

        # Drive Transceiver/CRG idle/reset from internal logic.
        self.sync += trx.tx_idle.eq(self.tx_idle)
        self.sync += crg.rx_reset.eq(self.rx_reset)
        self.sync += crg.tx_reset.eq(self.tx_reset)

        self.submodules.fsm = fsm = ResetInserter()(FSM(reset_state="RESET"))
        self.comb += fsm.reset.eq(retry_timer.done | align_timer.done)
        fsm.act("RESET",
            self.tx_idle.eq(1),
            trx.rx_cdrhold.eq(1),
            self.rx_reset.eq(1),
            self.tx_reset.eq(1),
            NextState("AWAIT-CRG-RESET")
        )
        fsm.act("AWAIT-CRG-RESET",
            self.tx_idle.eq(1),
            trx.rx_cdrhold.eq(1),
            NextValue(align_count, 4-1),
            If(trx.ready,
                # Set RX polarity to 0 (we don't know it at this point).
                NextValue(trx.rx_polarity, 0),
                # Alternate TX polarity on each retry.
                NextValue(trx.tx_polarity, ~trx.tx_polarity),
                NextState("COMINIT")
            )
        )
        fsm.act("COMINIT",
            self.tx_idle.eq(1),
            trx.rx_cdrhold.eq(1),
            trx.tx_cominit_stb.eq(1),
            If(trx.tx_cominit_ack & ~trx.rx_cominit_stb,
                NextState("AWAIT-COMINIT")
            )
        )
        fsm.act("AWAIT-COMINIT",
            self.tx_idle.eq(1),
            trx.rx_cdrhold.eq(1),
            retry_timer.wait.eq(1),
            If(trx.rx_cominit_stb,
                NextState("AWAIT-NO-COMINIT")
            )
        )
        fsm.act("AWAIT-NO-COMINIT",
            self.tx_idle.eq(1),
            trx.rx_cdrhold.eq(1),
            retry_timer.wait.eq(1),
            If(~trx.rx_cominit_stb,
                NextState("CALIBRATE")
            )
        )
        fsm.act("CALIBRATE",
            self.tx_idle.eq(1),
            trx.rx_cdrhold.eq(1),
            NextState("COMWAKE"),
        )
        fsm.act("COMWAKE",
            self.tx_idle.eq(1),
            trx.rx_cdrhold.eq(1),
            trx.tx_comwake_stb.eq(1),
            If(trx.tx_comwake_ack,
                NextState("AWAIT-COMWAKE")
            )
        )
        fsm.act("AWAIT-COMWAKE",
            self.tx_idle.eq(1),
            trx.rx_cdrhold.eq(1),
            retry_timer.wait.eq(1),
            If(trx.rx_comwake_stb,
                NextState("AWAIT-NO-COMWAKE")
            )
        )
        fsm.act("AWAIT-NO-COMWAKE",
            self.tx_idle.eq(1),
            trx.rx_cdrhold.eq(1),
            If(~trx.rx_comwake_stb,
                NextState("AWAIT-ALIGN")
            )
        )
        fsm.act("AWAIT-ALIGN",
            trx.rx_cdrhold.eq(1),
            source.data.eq(0x4a4a4a4a),  # D10.2
            source.charisk.eq(0b0000),
            align_timer.wait.eq(1),
            If(~trx.rx_idle,
                If(sink.valid & (self.sink.charisk == 0b0001) & (self.sink.data == primitives["ALIGN"]),
                    NextValue(trx.rx_polarity, 0),
                    NextState("SEND-ALIGN")
                ),
                If(sink.valid & (self.sink.charisk == 0b0001) & (self.sink.data == primitives["ALIGN_N"]),
                    NextValue(trx.rx_polarity, 1),
                    NextState("SEND-ALIGN")
                ),
            )
        )
        fsm.act("SEND-ALIGN",
            align_timer.wait.eq(1),
            source.data.eq(primitives["ALIGN"]),
            source.charisk.eq(0b0001),
            If(sink.valid & (sink.charisk == 0b0001),
                If(sink.data[0:8] == 0x7c,
                    NextValue(align_count, align_count - 1),
                ).Else(
                    NextValue(align_count, 4-1),
                )
            ),
            If(align_count == 0,
                NextState("READY")
            )
        )

        # Wait alignment stability for 5ms before declaring ctrl is ready, reset the RX part of
        # the transceiver when misalignment is detected.
        stability_timer = WaitTimer(5e-3*clk_freq)
        self.submodules += stability_timer

        fsm.act("READY",
            source.data.eq(primitives["SYNC"]),
            source.charisk.eq(0b0001),
            stability_timer.wait.eq(1),
            self.ready.eq(stability_timer.done),
            If(self.rx_idle,
                NextState("RESET"),
            ).Elif(self.misalign,
                self.rx_reset.eq(1),
                NextState("RESET_RX")
            )
        )
        fsm.act("RESET_RX",
            If(trx.ready,
                NextState("READY")
            )
        )

    def us(self, t):
        clk_period_us = 1e6/self.clk_freq
        return ceil(t/clk_period_us)



from litesata.common import *
from litesata.frontend.arbitration import LiteSATAArbiter, LiteSATACrossbar
from litesata.frontend.raid import LiteSATAStriping, LiteSATAMirroring
from litesata.frontend.bist import LiteSATABIST



#
# This file is part of LiteSATA.
#
# Copyright (c) 2015-2016 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

import unittest
from copy import deepcopy

from litesata.common import *
from litesata.core import LiteSATACore

from litex.soc.interconnect.stream_sim import *

from test.model.hdd import *


class CommandTXPacket(list):
    def __init__(self, write=0, read=0, sector=0, count=0, data=[]):
        self.ongoing = False
        self.done    = False
        self.write   = write
        self.read    = read
        self.sector  = sector
        self.count   = count
        for d in data:
            self.append(d)


class CommandStreamer(PacketStreamer):
    def __init__(self):
        self.source = stream.Endpoint(command_tx_description(32))

        # # #

        self.packets = []
        self.packet  = CommandTXPacket()
        self.packet.done = True

    def send(self, packet):
        packet = deepcopy(packet)
        self.packets.append(packet)
        return packet

    def send_blocking(self, packet):
        packet = self.send(packet)
        while not packet.done:
            yield

    @passive
    def generator(self):
        while True:
            if len(self.packets) and self.packet.done:
                self.packet = self.packets.pop(0)
            yield self.source.write.eq(self.packet.write)
            yield self.source.read.eq(self.packet.read)
            yield self.source.sector.eq(self.packet.sector)
            yield self.source.count.eq(self.packet.count)
            if not self.packet.ongoing and not self.packet.done:
                yield self.source.valid.eq(1)
                if len(self.packet) > 0:
                    yield self.source.data.eq(self.packet.pop(0))
                self.packet.ongoing = True
            elif (yield self.source.valid) and (yield self.source.ready):
                yield self.source.last.eq(len(self.packet) == 1)
                if len(self.packet) > 0:
                    yield self.source.valid.eq(1)
                    yield self.source.data.eq(self.packet.pop(0))
                else:
                    self.packet.done = True
                    yield self.source.valid.eq(0)

            yield


class CommandRXPacket(list):
    def __init__(self):
        self.ongoing = False
        self.done    = False
        self.write   = 0
        self.read    = 0
        self.failed  = 0


class CommandLogger(PacketLogger):
    def __init__(self):
        PacketLogger.__init__(self, command_rx_description(32), CommandRXPacket)
        self.first = True

    @passive
    def generator(self):
        while True:
            yield self.sink.ready.eq(1)
            if (yield self.sink.valid) and self.first:
                self.packet = CommandRXPacket()
                self.packet.write  = (yield self.sink.write)
                self.packet.read   = (yield self.sink.read)
                self.packet.failed = (yield self.sink.failed)
                self.packet.append((yield self.sink.data))
                self.first = False
            elif (yield self.sink.valid):
                self.packet.append((yield self.sink.data))
            if (yield self.sink.valid) and (yield self.sink.last):
                self.packet.done = True
                self.first = True
            yield


class TestCommand(unittest.TestCase):
    def test_command(self):
        def generator(dut):
            hdd = dut.hdd
            hdd.malloc(0, 64)
            write_data   = [i for i in range(sectors2dwords(2))]
            write_len    = dwords2sectors(len(write_data))
            write_packet = CommandTXPacket(write=1, sector=2, count=write_len, data=write_data)
            yield from dut.streamer.send_blocking(write_packet)
            yield from dut.logger.receive()

            read_packet = CommandTXPacket(read=1, sector=2, count=write_len)
            yield from dut.streamer.send_blocking(read_packet)
            yield from dut.logger.receive()
            read_data = dut.logger.packet

            # Check results
            s, l, e = check(write_data, read_data)
            print("shift " + str(s) + " / length " + str(l) + " / errors " + str(e))
            self.assertEqual(s, 0)
            self.assertEqual(e, 0)

        class DUT(Module):
            def __init__(self):
                self.submodules.hdd = HDD(
                    link_debug         = False,
                    link_random_level  = 50,
                    transport_debug    = False,
                    transport_loopback = False,
                    hdd_debug          = True)
                self.submodules.core = LiteSATACore(self.hdd.phy)

                self.submodules.streamer = CommandStreamer()
                self.submodules.streamer_randomizer = Randomizer(command_tx_description(32), level=50)

                self.submodules.logger = CommandLogger()
                self.submodules.logger_randomizer = Randomizer(command_rx_description(32), level=50)

                self.submodules.pipeline = Pipeline(
                    self.streamer,
                    self.streamer_randomizer,
                    self.core,
                    self.logger_randomizer,
                    self.logger
                )

        dut = DUT()
        generators = {
            "sys" :   [generator(dut),
                       dut.hdd.link.generator(),
                       dut.streamer.generator(),
                       dut.streamer_randomizer.generator(),
                       dut.logger.generator(),
                       dut.logger_randomizer.generator(),
                       dut.hdd.phy.rx.generator(),
                       dut.hdd.phy.tx.generator()]
        }
        clocks = {"sys": 10}
        run_simulation(dut, generators, clocks)



#!/usr/bin/env python3

#
# This file is part of LiteSATA.
#
# Copyright (c) 2014-2020 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

"""
LiteSATA standalone core generator

LiteSATA aims to be directly used as a python package when the SoC is created using LiteX. However,
for some use cases it could be interesting to generate a standalone verilog file of the core:
- integration of the core in a SoC using a more traditional flow.
- need to version/package the core.
- avoid Migen/LiteX dependencies.
- etc...

The standalone core is generated from a YAML configuration file that allows the user to generate
easily a custom configuration of the core.

Current version of the generator is limited to:
- Xilinx 7-Series.
"""

import yaml
import argparse
import subprocess

from migen import *
from migen.genlib.resetsync import AsyncResetSynchronizer

from litex.soc.integration.soc_core import *
from litex.soc.integration.builder import *

from litesata.phy import LiteSATAPHY
from litesata.core import LiteSATACore
from litesata.frontend.arbitration import LiteSATACrossbar

from litex.build.generic_platform import *

# IOs/Interfaces -----------------------------------------------------------------------------------

def get_clkin_ios():
    return [
        ("clk", 0, Pins(1)),
        ("rst", 0, Pins(1))
    ]

def get_sata_ios():
    return [
        ("sata", 0,
            Subsignal("clk_p", Pins(1)),
            Subsignal("clk_n", Pins(1)),
            Subsignal("rx_p",  Pins(1)),
            Subsignal("rx_n",  Pins(1)),
            Subsignal("tx_p",  Pins(1)),
            Subsignal("tx_n",  Pins(1)),
        ),
    ]

def get_ctrl_ios():
    return [
        ("enable", 0, Pins(1)),
        ("ready",  0, Pins(1)),
    ]

def get_native_user_port_ios(_id):
    return [
        ("user_port_{}".format(_id), 0,
            Subsignal("sink_valid",    Pins(1)),
            Subsignal("sink_last",     Pins(1)),
            Subsignal("sink_ready",    Pins(1)),
            Subsignal("sink_write",    Pins(1)),
            Subsignal("sink_read",     Pins(1)),
            Subsignal("sink_identify", Pins(1)),
            Subsignal("sink_sector",   Pins(48)),
            Subsignal("sink_count",    Pins(16)),
            Subsignal("sink_data",     Pins(32)),

            Subsignal("source_valid",    Pins(1)),
            Subsignal("source_last",     Pins(1)),
            Subsignal("source_ready",    Pins(1)),
            Subsignal("source_write",    Pins(1)),
            Subsignal("source_read",     Pins(1)),
            Subsignal("source_identify", Pins(1)),
            Subsignal("source_end",      Pins(1)),
            Subsignal("source_failed",   Pins(1)),
            Subsignal("source_data",     Pins(32)),
        ),
    ]

# CRG ----------------------------------------------------------------------------------------------

class LiteSATACRG(Module):
    def __init__(self, platform, sys_clk_freq):
        self.clock_domains.cd_sys = ClockDomain()

        # # #

        platform.add_extension(get_clkin_ios())
        self.comb += self.cd_sys.clk.eq(platform.request("clk"))
        self.specials += AsyncResetSynchronizer(self.cd_sys, platform.request("rst"))

# Core ---------------------------------------------------------------------------------------------

class _LiteSATACore(SoCMini):
    SoCMini.mem_map["csr"] = 0x00000000
    SoCMini.csr_map = {
        "ctrl":           0,
    }
    def __init__(self, platform, core_config):
        platform.add_extension(get_sata_ios())
        platform.add_extension(get_ctrl_ios())
        sys_clk_freq = int(float(core_config.get("clk_freq")))
        sata_clk_freq = {"gen1": 75e6, "gen2": 150e6, "gen3": 300e6}[core_config["gen"]]

        # SoCMini ----------------------------------------------------------------------------------
        SoCMini.__init__(self, platform, clk_freq=sys_clk_freq,
            csr_data_width = 32,
            ident          = "LiteSATA standalone core",
            ident_version  = True
        )

        # CRG --------------------------------------------------------------------------------------
        self.submodules.crg = LiteSATACRG(platform, sys_clk_freq)

        # SATA PHY ---------------------------------------------------------------------------------
        self.submodules.sata_phy = LiteSATAPHY(platform.device,
            pads       = platform.request("sata"),
            gen        = core_config["gen"],
            clk_freq   = sys_clk_freq,
            data_width = 16,
            with_csr   = False)
        self.comb += self.sata_phy.enable.eq(platform.request("enable"))
        self.comb += platform.request("ready").eq(self.sata_phy.ready)

        # SATA Core --------------------------------------------------------------------------------
        self.submodules.sata_core = LiteSATACore(self.sata_phy)

        # SATA Crossbar ----------------------------------------------------------------------------
        self.submodules.sata_crossbar = LiteSATACrossbar(self.sata_core)

        # SATA User Ports --------------------------------------------------------------------------
        for name, port in core_config["user_ports"].items():
            # Native
            if port["type"] == "native":
                assert port["data_width"] == 32
                user_port = self.sata_crossbar.get_port()
                platform.add_extension(get_native_user_port_ios(name))
                user_port_pads = platform.request("user_port_{}".format(name))

                self.comb += [
                    user_port.sink.valid.eq(user_port_pads.sink_valid),
                    user_port.sink.last.eq(user_port_pads.sink_last),
                    user_port.sink.write.eq(user_port_pads.sink_write),
                    user_port.sink.read.eq(user_port_pads.sink_read),
                    user_port.sink.identify.eq(user_port_pads.sink_identify),
                    user_port.sink.sector.eq(user_port_pads.sink_sector),
                    user_port.sink.count.eq(user_port_pads.sink_count),

                    user_port_pads.sink_ready.eq(user_port.sink.ready),
                ]

                self.comb += [
                    user_port_pads.source_valid.eq(user_port.source.valid),
                    user_port_pads.source_last.eq(user_port.source.last),
                    user_port_pads.source_write.eq(user_port.source.write),
                    user_port_pads.source_read.eq(user_port.source.read),
                    user_port_pads.source_identify.eq(user_port.source.identify),
                    user_port_pads.source_end.eq(user_port.source.end),
                    user_port_pads.source_failed.eq(user_port.source.failed),
                    user_port_pads.source_data.eq(user_port.source.data),

                    user_port.source.ready.eq(user_port_pads.source_ready),
                ]
            else:
                raise ValueError("Unsupported port type: {}".format(port["type"]))


        # Timing constraints
        platform.add_period_constraint(self.sata_phy.crg.cd_sata_tx.clk, 1e9/sata_clk_freq)
        platform.add_period_constraint(self.sata_phy.crg.cd_sata_rx.clk, 1e9/sata_clk_freq)
        self.platform.add_false_path_constraints(
            self.crg.cd_sys.clk,
            self.sata_phy.crg.cd_sata_tx.clk,
            self.sata_phy.crg.cd_sata_rx.clk)

# Build --------------------------------------------------------------------------------------------

def main():
    parser = argparse.ArgumentParser(description="LiteSATA standalone core generator")
    parser.add_argument("config", help="YAML config file")
    args = parser.parse_args()
    core_config = yaml.load(open(args.config).read(), Loader=yaml.Loader)

    # Convert YAML elements to Python/LiteX --------------------------------------------------------
    for k, v in core_config.items():
        replaces = {"False": False, "True": True, "None": None}
        for r in replaces.keys():
            if v == r:
                core_config[k] = replaces[r]

    # Generate core --------------------------------------------------------------------------------
    if core_config["phy"] == "A7SATAPHY":
        from litex.build.xilinx import XilinxPlatform
        platform = XilinxPlatform("xc7a", io=[])
    elif core_config["phy"] == "K7SATAPHY":
        from litex.build.xilinx import XilinxPlatform
        platform = XilinxPlatform("xc7k", io=[])
    else:
        raise ValueError("Unsupported SATA PHY: {}".format(core_config["phy"]))
    soc      = _LiteSATACore(platform, core_config)
    builder  = Builder(soc, output_dir="build", compile_gateware=False)
    builder.build(build_name="litesata_core", regular_comb=True)

if __name__ == "__main__":
    main()



#!/usr/bin/env python3

from setuptools import setup
from setuptools import find_packages

with open("README.md", "r", encoding="utf-8") as fp:
    long_description = fp.read()

setup(
    name="litesata",
    description="Small footprint and configurable SATA core",
    long_description              = long_description,
    long_description_content_type = "text/markdown",
    version="2023.08",
    author="Florent Kermarrec",
    author_email="florent@enjoy-digital.fr",
    url="http://enjoy-digital.fr",
    download_url="https://github.com/enjoy-digital/litesata",
    test_suite="test",
    license="BSD",
    python_requires="~=3.6",
    packages=find_packages(exclude=("test*", "sim*", "doc*", "examples*")),
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "litesata_gen=litesata.gen:main",
        ],
    },
)



#
# This file is part of LiteSATA.
#
# Copyright (c) 2015-2016 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

import unittest

from litesata.common import *
from litesata.core import LiteSATACore
from litesata.frontend.arbitration import LiteSATACrossbar
from litesata.frontend.bist import LiteSATABISTGenerator, LiteSATABISTChecker
from litesata.frontend.raid import LiteSATAMirroring

from litex.soc.interconnect.stream_sim import *

from test.model.hdd import *


class TestMirroring(unittest.TestCase):
    def test_mirroring(self):
        def generator(dut):
            dut.hdd0.malloc(0, 64)
            dut.hdd1.malloc(0, 64)
            sector = 0
            count = 1

            for i in range(2):
                for dut_generator in [dut.generator0, dut.generator1]:
                    # Write data (alternate generators)
                    yield dut_generator.sector.eq(sector)
                    yield dut_generator.count.eq(count)
                    yield dut_generator.start.eq(1)
                    yield
                    yield dut_generator.start.eq(0)
                    yield
                    while not (yield dut_generator.done):
                        yield

                    # Verify data on the 2 hdds in //
                    yield dut.checker0.sector.eq(sector)
                    yield dut.checker0.count.eq(count)
                    yield dut.checker0.start.eq(1)
                    yield dut.checker1.sector.eq(sector)
                    yield dut.checker1.count.eq(count)
                    yield dut.checker1.start.eq(1)
                    yield
                    yield dut.checker0.start.eq(0)
                    yield dut.checker1.start.eq(0)
                    yield
                    while not (yield dut.checker0.done) or not (yield dut.checker1.done):
                        yield
                    errors = (yield dut.checker0.errors) + (yield dut.checker1.errors)
                    print("errors {}".format(errors))
                    self.assertEqual(errors, 0)

                    # Prepare next iteration
                    sector += 1

        class DUT(Module):
            def __init__(self):
                self.submodules.hdd0 = HDD(n=0,
                        link_debug         = False,
                        link_random_level  = 0,
                        transport_debug    = False,
                        transport_loopback = False,
                        hdd_debug          = True)
                self.submodules.core0 = LiteSATACore(self.hdd0.phy)

                self.submodules.hdd1 = HDD(n=1,
                        link_debug         = False,
                        link_random_level  = 0,
                        transport_debug    = False,
                        transport_loopback = False,
                        hdd_debug          = True)
                self.submodules.core1 = LiteSATACore(self.hdd1.phy)

                self.submodules.mirroring = LiteSATAMirroring([self.core0, self.core1])

                self.submodules.crossbar0  = LiteSATACrossbar(self.mirroring.ports[0])
                self.submodules.generator0 = LiteSATABISTGenerator(self.crossbar0.get_port())
                self.submodules.checker0   = LiteSATABISTChecker(self.crossbar0.get_port())

                self.submodules.crossbar1  = LiteSATACrossbar(self.mirroring.ports[1])
                self.submodules.generator1 = LiteSATABISTGenerator(self.crossbar1.get_port())
                self.submodules.checker1   = LiteSATABISTChecker(self.crossbar1.get_port())

        dut = DUT()
        generators = {
            "sys" :   [generator(dut),
                       dut.hdd0.link.generator(),
                       dut.hdd0.phy.rx.generator(),
                       dut.hdd0.phy.tx.generator(),
                       dut.hdd1.link.generator(),
                       dut.hdd1.phy.rx.generator(),
                       dut.hdd1.phy.tx.generator()]
        }
        clocks = {"sys": 10}
        run_simulation(dut, generators, clocks)



import re

from litesata.common import *
from litesata.phy.ctrl import *
from litesata.phy.datapath import *


class LiteSATAPHY(Module, AutoCSR):
    """SATA PHY

    Manages the low level interface between the SATA core and the device.

    This modules use FPGA transceivers (high speed serializers/deserializer) to
    communicates with the SATA devices. Since transceivers are primitives inside
    the FPGA, device passed as parameter is used to select the right PHY. PHY is
    composed of 3 main modules:
    - Transceiver and clocking (vendor specific)
    - Control (vendor agnostic)
    - Datapath (vendor agnostic)

    For now, the Kintex7/Zynq(with PL based on K7) PHY is the only one available,
    but the achitecture is modular enough to accept others PHYs.
    """
    def __init__(self, device, pads, gen, clk_freq, refclk=None, data_width=16, qpll=None, with_csr=True):
        self.pads   = pads
        self.gen    = gen
        self.refclk = refclk

        # Control/Status
        self.enable = Signal()
        self.ready  = Signal()

        # Transceiver / Clocks

        # Kintex7
        if re.match("^xc7k", device):
            from litesata.phy.k7sataphy import K7LiteSATAPHYCRG, K7LiteSATAPHY
            self.submodules.phy = K7LiteSATAPHY(pads, gen, clk_freq, data_width)
            self.submodules.crg = K7LiteSATAPHYCRG(refclk, pads, self.phy, gen)

        # Artix7
        elif re.match("^xc7a", device):
            from litesata.phy.a7sataphy import A7LiteSATAPHYCRG, A7LiteSATAPHY
            self.submodules.phy = A7LiteSATAPHY(pads, gen, clk_freq, data_width, tx_buffer_enable=True, qpll=qpll)
            self.submodules.crg = A7LiteSATAPHYCRG(refclk, pads, self.phy, gen,  tx_buffer_enable=True)

        # Kintex/Virtex Ultrascale
        elif re.match("^xc[kv]u[0-9]+-", device):
            from litesata.phy.ussataphy import USLiteSATAPHYCRG, USLiteSATAPHY
            self.submodules.phy = USLiteSATAPHY(pads, gen, clk_freq, data_width)
            self.submodules.crg = USLiteSATAPHYCRG(refclk, pads, self.phy, gen)

        # Kintex/Virtex Ultrascale+
        elif re.match("^xc[kv]u[0-9]+p-", device):
            from litesata.phy.uspsataphy import USPLiteSATAPHYCRG, USPLiteSATAPHY
            self.submodules.phy = USPLiteSATAPHY(pads, gen, clk_freq, data_width)
            self.submodules.crg = USPLiteSATAPHYCRG(refclk, pads, self.phy, gen)

        # ECP5
        elif re.match("^LFE5UM5G-", device):
            from litesata.phy.ecp5sataphy import ECP5LiteSATAPHYCRG, ECP5LiteSATAPHY
            self.submodules.phy = ECP5LiteSATAPHY(refclk, pads, gen, clk_freq, data_width)
            self.submodules.crg = ECP5LiteSATAPHYCRG(self.phy)

        # Unknown
        else:
            raise NotImplementedError(f"Unsupported {device} Device.")

        # Control
        self.submodules.ctrl = LiteSATAPHYCtrl(self.phy, self.crg, clk_freq)

        # Datapath
        self.submodules.datapath = LiteSATAPHYDatapath(self.phy, self.ctrl)
        self.comb += [
            self.ctrl.rx_idle.eq(self.datapath.rx_idle),
            self.ctrl.misalign.eq(self.datapath.misalign)
        ]
        self.sink, self.source = self.datapath.sink, self.datapath.source

        # Restart/Status
        if hasattr(self.phy, "tx_init") and hasattr(self.phy, "rx_init"):
            self.comb += self.phy.tx_init.restart.eq(~self.enable)
            self.comb += self.phy.rx_init.restart.eq(~self.enable | self.ctrl.rx_reset)
        self.comb += self.ready.eq(self.phy.ready & self.ctrl.ready)

        # CSRs
        if with_csr:
            self.add_csr()

    def  add_csr(self):
        self._enable = CSRStorage(reset=1)
        self._status = CSRStatus(fields=[
            CSRField("ready", size=1, values=[
                    ("``0b0``", "PHY not initialized."),
                    ("``0b1``", "PHY initialized and ready.")
            ]),
            CSRField("tx_ready", size=1, values=[
                    ("``0b0``", "TX not initialized."),
                    ("``0b1``", "TX initialized and ready.")
            ]),
            CSRField("rx_ready", size=1, values=[
                    ("``0b0``", "RX not initialized."),
                    ("``0b1``", "RX initialized and ready.")
            ]),
            CSRField("ctrl_ready", size=1, values=[
                    ("``0b0``", "Ctrl/OOB not initialized."),
                    ("``0b1``", "Ctrl/OOB initialized and ready.")
            ]),
        ])

        self.comb += self.enable.eq(self._enable.storage)
        self.comb += self._status.fields.ready.eq(self.phy.ready & self.ctrl.ready)
        if hasattr(self.phy, "tx_init") and hasattr(self.phy, "rx_init"):
            self.comb += self._status.fields.tx_ready.eq(self.phy.tx_init.done)
            self.comb += self._status.fields.rx_ready.eq(self.phy.rx_init.done)
        else:
            self.comb += self._status.fields.tx_ready.eq(self.phy.ready)
            self.comb += self._status.fields.rx_ready.eq(self.phy.ready)
        self.comb += self._status.fields.ctrl_ready.eq(self.ctrl.ready)



#
# This file is part of LiteSATA.
#
# Copyright (c) 2015-2020 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

from litesata.common import *

from litex.soc.interconnect.csr import *

# LiteSATAIdentify ---------------------------------------------------------------------------------

class LiteSATAIdentify(Module):
    def __init__(self, user_port):
        self.start      = Signal()
        self.done       = Signal()
        self.data_width = user_port.dw

        fifo = ResetInserter()(stream.SyncFIFO([("data", 32)], 512, buffered=True))
        self.submodules += fifo
        self.source = fifo.source

        # # #

        source, sink = user_port.sink, user_port.source

        self.submodules.fsm = fsm = FSM(reset_state="IDLE")
        fsm.act("IDLE",
            self.done.eq(1),
            If(self.start,
                NextState("SEND-CMD")
            )
        )
        self.comb += [
            source.last.eq(1),
            source.identify.eq(1),
        ]
        fsm.act("SEND-CMD",
            fifo.reset.eq(1),
            source.valid.eq(1),
            If(source.valid & source.ready,
                NextState("WAIT-ACK")
            )
        )
        fsm.act("WAIT-ACK",
            If(sink.valid & sink.identify,
                NextState("RECEIVE-DATA")
            )
        )
        self.comb += fifo.sink.data.eq(sink.data)
        fsm.act("RECEIVE-DATA",
            sink.ready.eq(fifo.sink.ready),
            If(sink.valid,
                fifo.sink.valid.eq(1),
                If(sink.last,
                    NextState("IDLE")
                )
            )
        )

# LiteSATAIdentifyCSR ------------------------------------------------------------------------------

class LiteSATAIdentifyCSR(Module, AutoCSR):
    def __init__(self, bist_identify):
        self._start        = CSR()
        self._done         = CSRStatus()
        self._data_width   = CSRStatus(16, reset=bist_identify.data_width)
        self._source_valid = CSRStatus()
        self._source_ready = CSR()
        self._source_data  = CSRStatus(32)

        # # #

        self.submodules += bist_identify
        self.comb += [
            bist_identify.start.eq(self._start.r & self._start.re),
            self._done.status.eq(bist_identify.done),

            self._source_valid.status.eq(bist_identify.source.valid),
            self._source_data.status.eq(bist_identify.source.data),
            bist_identify.source.ready.eq(self._source_ready.r & self._source_ready.re)
        ]
